# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model_registry.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from . import databricks_pb2 as databricks__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14model_registry.proto\x12\x06mlflow\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"\x89\x02\n\x0fRegisteredModel\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x02 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x03 \x01(\x03\x12\x0f\n\x07user_id\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12-\n\x0flatest_versions\x18\x06 \x03(\x0b\x32\x14.mlflow.ModelVersion\x12(\n\x04tags\x18\x07 \x03(\x0b\x32\x1a.mlflow.RegisteredModelTag\x12-\n\x07\x61liases\x18\x08 \x03(\x0b\x32\x1c.mlflow.RegisteredModelAlias\"\xd4\x02\n\x0cModelVersion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x03 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x04 \x01(\x03\x12\x0f\n\x07user_id\x18\x05 \x01(\t\x12\x15\n\rcurrent_stage\x18\x06 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x07 \x01(\t\x12\x0e\n\x06source\x18\x08 \x01(\t\x12\x0e\n\x06run_id\x18\t \x01(\t\x12*\n\x06status\x18\n \x01(\x0e\x32\x1a.mlflow.ModelVersionStatus\x12\x16\n\x0estatus_message\x18\x0b \x01(\t\x12%\n\x04tags\x18\x0c \x03(\x0b\x32\x17.mlflow.ModelVersionTag\x12\x10\n\x08run_link\x18\r \x01(\t\x12\x0f\n\x07\x61liases\x18\x0e \x03(\t\"\xd6\x01\n\x15\x43reateRegisteredModel\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12(\n\x04tags\x18\x02 \x03(\x0b\x32\x1a.mlflow.RegisteredModelTag\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x1a=\n\x08Response\x12\x31\n\x10registered_model\x18\x01 \x01(\x0b\x32\x17.mlflow.RegisteredModel:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xa9\x01\n\x15RenameRegisteredModel\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x10\n\x08new_name\x18\x02 \x01(\t\x1a=\n\x08Response\x12\x31\n\x10registered_model\x18\x01 \x01(\x0b\x32\x17.mlflow.RegisteredModel:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xac\x01\n\x15UpdateRegisteredModel\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x1a=\n\x08Response\x12\x31\n\x10registered_model\x18\x01 \x01(\x0b\x32\x17.mlflow.RegisteredModel:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"d\n\x15\x44\x65leteRegisteredModel\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x94\x01\n\x12GetRegisteredModel\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a=\n\x08Response\x12\x31\n\x10registered_model\x18\x01 \x01(\x0b\x32\x17.mlflow.RegisteredModel:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xee\x01\n\x16SearchRegisteredModels\x12\x0e\n\x06\x66ilter\x18\x01 \x01(\t\x12\x18\n\x0bmax_results\x18\x02 \x01(\x03:\x03\x31\x30\x30\x12\x10\n\x08order_by\x18\x03 \x03(\t\x12\x12\n\npage_token\x18\x04 \x01(\t\x1aW\n\x08Response\x12\x32\n\x11registered_models\x18\x01 \x03(\x0b\x32\x17.mlflow.RegisteredModel\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x9e\x01\n\x11GetLatestVersions\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0e\n\x06stages\x18\x02 \x03(\t\x1a\x38\n\x08Response\x12,\n\x0emodel_versions\x18\x01 \x03(\x0b\x32\x14.mlflow.ModelVersion:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x82\x02\n\x12\x43reateModelVersion\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x14\n\x06source\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0e\n\x06run_id\x18\x03 \x01(\t\x12%\n\x04tags\x18\x04 \x03(\x0b\x32\x17.mlflow.ModelVersionTag\x12\x10\n\x08run_link\x18\x05 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x1a\x37\n\x08Response\x12+\n\rmodel_version\x18\x01 \x01(\x0b\x32\x14.mlflow.ModelVersion:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xba\x01\n\x12UpdateModelVersion\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x1a\x37\n\x08Response\x12+\n\rmodel_version\x18\x01 \x01(\x0b\x32\x14.mlflow.ModelVersion:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xec\x01\n\x1bTransitionModelVersionStage\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05stage\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x12\'\n\x19\x61rchive_existing_versions\x18\x04 \x01(\x08\x42\x04\xf8\x86\x19\x01\x1a\x37\n\x08Response\x12+\n\rmodel_version\x18\x01 \x01(\x0b\x32\x14.mlflow.ModelVersion:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"x\n\x12\x44\x65leteModelVersion\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xa2\x01\n\x0fGetModelVersion\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a\x37\n\x08Response\x12+\n\rmodel_version\x18\x01 \x01(\x0b\x32\x14.mlflow.ModelVersion:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xe8\x01\n\x13SearchModelVersions\x12\x0e\n\x06\x66ilter\x18\x01 \x01(\t\x12\x1b\n\x0bmax_results\x18\x02 \x01(\x03:\x06\x32\x30\x30\x30\x30\x30\x12\x10\n\x08order_by\x18\x03 \x03(\t\x12\x12\n\npage_token\x18\x04 \x01(\t\x1aQ\n\x08Response\x12,\n\x0emodel_versions\x18\x01 \x03(\x0b\x32\x14.mlflow.ModelVersion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x96\x01\n\x1aGetModelVersionDownloadUri\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a \n\x08Response\x12\x14\n\x0c\x61rtifact_uri\x18\x01 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"-\n\x0fModelVersionTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"0\n\x12RegisteredModelTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x8c\x01\n\x15SetRegisteredModelTag\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xa0\x01\n\x12SetModelVersionTag\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x11\n\x03key\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x04 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"z\n\x18\x44\x65leteRegisteredModelTag\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x8e\x01\n\x15\x44\x65leteModelVersionTag\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x11\n\x03key\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"6\n\x14RegisteredModelAlias\x12\r\n\x05\x61lias\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"\x92\x01\n\x17SetRegisteredModelAlias\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05\x61lias\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"~\n\x1a\x44\x65leteRegisteredModelAlias\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05\x61lias\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xa7\x01\n\x16GetModelVersionByAlias\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05\x61lias\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a\x37\n\x08Response\x12+\n\rmodel_version\x18\x01 \x01(\x0b\x32\x14.mlflow.ModelVersion:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]*R\n\x12ModelVersionStatus\x12\x18\n\x14PENDING_REGISTRATION\x10\x01\x12\x17\n\x13\x46\x41ILED_REGISTRATION\x10\x02\x12\t\n\x05READY\x10\x03\x32\x80\x1e\n\x14ModelRegistryService\x12\xae\x01\n\x15\x63reateRegisteredModel\x12\x1d.mlflow.CreateRegisteredModel\x1a&.mlflow.CreateRegisteredModel.Response\"N\xf2\x86\x19J\n.\n\x04POST\x12 /mlflow/registered-models/create\x1a\x04\x08\x02\x10\x00\x10\x01*\x16\x43reate RegisteredModel\x12\xae\x01\n\x15renameRegisteredModel\x12\x1d.mlflow.RenameRegisteredModel\x1a&.mlflow.RenameRegisteredModel.Response\"N\xf2\x86\x19J\n.\n\x04POST\x12 /mlflow/registered-models/rename\x1a\x04\x08\x02\x10\x00\x10\x01*\x16Rename RegisteredModel\x12\xaf\x01\n\x15updateRegisteredModel\x12\x1d.mlflow.UpdateRegisteredModel\x1a&.mlflow.UpdateRegisteredModel.Response\"O\xf2\x86\x19K\n/\n\x05PATCH\x12 /mlflow/registered-models/update\x1a\x04\x08\x02\x10\x00\x10\x01*\x16Update RegisteredModel\x12\xb0\x01\n\x15\x64\x65leteRegisteredModel\x12\x1d.mlflow.DeleteRegisteredModel\x1a&.mlflow.DeleteRegisteredModel.Response\"P\xf2\x86\x19L\n0\n\x06\x44\x45LETE\x12 /mlflow/registered-models/delete\x1a\x04\x08\x02\x10\x00\x10\x01*\x16\x44\x65lete RegisteredModel\x12\x9e\x01\n\x12getRegisteredModel\x12\x1a.mlflow.GetRegisteredModel\x1a#.mlflow.GetRegisteredModel.Response\"G\xf2\x86\x19\x43\n*\n\x03GET\x12\x1d/mlflow/registered-models/get\x1a\x04\x08\x02\x10\x00\x10\x01*\x13Get RegisteredModel\x12\xb1\x01\n\x16searchRegisteredModels\x12\x1e.mlflow.SearchRegisteredModels\x1a\'.mlflow.SearchRegisteredModels.Response\"N\xf2\x86\x19J\n-\n\x03GET\x12 /mlflow/registered-models/search\x1a\x04\x08\x02\x10\x00\x10\x01*\x17Search RegisteredModels\x12\xef\x01\n\x11getLatestVersions\x12\x19.mlflow.GetLatestVersions\x1a\".mlflow.GetLatestVersions.Response\"\x9a\x01\xf2\x86\x19\x95\x01\n;\n\x04POST\x12-/mlflow/registered-models/get-latest-versions\x1a\x04\x08\x02\x10\x00\n:\n\x03GET\x12-/mlflow/registered-models/get-latest-versions\x1a\x04\x08\x02\x10\x00\x10\x01*\x18Get Latest ModelVersions\x12\x9f\x01\n\x12\x63reateModelVersion\x12\x1a.mlflow.CreateModelVersion\x1a#.mlflow.CreateModelVersion.Response\"H\xf2\x86\x19\x44\n+\n\x04POST\x12\x1d/mlflow/model-versions/create\x1a\x04\x08\x02\x10\x00\x10\x01*\x13\x43reate ModelVersion\x12\xa0\x01\n\x12updateModelVersion\x12\x1a.mlflow.UpdateModelVersion\x1a#.mlflow.UpdateModelVersion.Response\"I\xf2\x86\x19\x45\n,\n\x05PATCH\x12\x1d/mlflow/model-versions/update\x1a\x04\x08\x02\x10\x00\x10\x01*\x13Update ModelVersion\x12\xce\x01\n\x1btransitionModelVersionStage\x12#.mlflow.TransitionModelVersionStage\x1a,.mlflow.TransitionModelVersionStage.Response\"\\\xf2\x86\x19X\n5\n\x04POST\x12\'/mlflow/model-versions/transition-stage\x1a\x04\x08\x02\x10\x00\x10\x01*\x1dTransition ModelVersion Stage\x12\xa1\x01\n\x12\x64\x65leteModelVersion\x12\x1a.mlflow.DeleteModelVersion\x1a#.mlflow.DeleteModelVersion.Response\"J\xf2\x86\x19\x46\n-\n\x06\x44\x45LETE\x12\x1d/mlflow/model-versions/delete\x1a\x04\x08\x02\x10\x00\x10\x01*\x13\x44\x65lete ModelVersion\x12\x8f\x01\n\x0fgetModelVersion\x12\x17.mlflow.GetModelVersion\x1a .mlflow.GetModelVersion.Response\"A\xf2\x86\x19=\n\'\n\x03GET\x12\x1a/mlflow/model-versions/get\x1a\x04\x08\x02\x10\x00\x10\x01*\x10Get ModelVersion\x12\xa6\x01\n\x13searchModelVersions\x12\x1b.mlflow.SearchModelVersions\x1a$.mlflow.SearchModelVersions.Response\"L\xf2\x86\x19\x44\n*\n\x03GET\x12\x1d/mlflow/model-versions/search\x1a\x04\x08\x02\x10\x00\x10\x01*\x14Search ModelVersions\xba\x8c\x19\x00\x12\xd8\x01\n\x1agetModelVersionDownloadUri\x12\".mlflow.GetModelVersionDownloadUri\x1a+.mlflow.GetModelVersionDownloadUri.Response\"i\xf2\x86\x19\x65\n4\n\x03GET\x12\'/mlflow/model-versions/get-download-uri\x1a\x04\x08\x02\x10\x00\x10\x01*+Get Download URI For ModelVersion Artifacts\x12\xb1\x01\n\x15setRegisteredModelTag\x12\x1d.mlflow.SetRegisteredModelTag\x1a&.mlflow.SetRegisteredModelTag.Response\"Q\xf2\x86\x19M\n/\n\x04POST\x12!/mlflow/registered-models/set-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\x18Set Registered Model Tag\x12\xa2\x01\n\x12setModelVersionTag\x12\x1a.mlflow.SetModelVersionTag\x1a#.mlflow.SetModelVersionTag.Response\"K\xf2\x86\x19G\n,\n\x04POST\x12\x1e/mlflow/model-versions/set-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\x15Set Model Version Tag\x12\xc2\x01\n\x18\x64\x65leteRegisteredModelTag\x12 .mlflow.DeleteRegisteredModelTag\x1a).mlflow.DeleteRegisteredModelTag.Response\"Y\xf2\x86\x19U\n4\n\x06\x44\x45LETE\x12$/mlflow/registered-models/delete-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\x1b\x44\x65lete Registered Model Tag\x12\xb3\x01\n\x15\x64\x65leteModelVersionTag\x12\x1d.mlflow.DeleteModelVersionTag\x1a&.mlflow.DeleteModelVersionTag.Response\"S\xf2\x86\x19O\n1\n\x06\x44\x45LETE\x12!/mlflow/model-versions/delete-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\x18\x44\x65lete Model Version Tag\x12\xb7\x01\n\x17setRegisteredModelAlias\x12\x1f.mlflow.SetRegisteredModelAlias\x1a(.mlflow.SetRegisteredModelAlias.Response\"Q\xf2\x86\x19M\n-\n\x04POST\x12\x1f/mlflow/registered-models/alias\x1a\x04\x08\x02\x10\x00\x10\x01*\x1aSet Registered Model Alias\x12\xc5\x01\n\x1a\x64\x65leteRegisteredModelAlias\x12\".mlflow.DeleteRegisteredModelAlias\x1a+.mlflow.DeleteRegisteredModelAlias.Response\"V\xf2\x86\x19R\n/\n\x06\x44\x45LETE\x12\x1f/mlflow/registered-models/alias\x1a\x04\x08\x02\x10\x00\x10\x01*\x1d\x44\x65lete Registered Model Alias\x12\xb3\x01\n\x16getModelVersionByAlias\x12\x1e.mlflow.GetModelVersionByAlias\x1a\'.mlflow.GetModelVersionByAlias.Response\"P\xf2\x86\x19L\n,\n\x03GET\x12\x1f/mlflow/registered-models/alias\x1a\x04\x08\x02\x10\x00\x10\x01*\x1aGet Model Version by AliasB!\n\x14org.mlflow.api.proto\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

_MODELVERSIONSTATUS = DESCRIPTOR.enum_types_by_name['ModelVersionStatus']
ModelVersionStatus = enum_type_wrapper.EnumTypeWrapper(_MODELVERSIONSTATUS)
PENDING_REGISTRATION = 1
FAILED_REGISTRATION = 2
READY = 3


_REGISTEREDMODEL = DESCRIPTOR.message_types_by_name['RegisteredModel']
_MODELVERSION = DESCRIPTOR.message_types_by_name['ModelVersion']
_CREATEREGISTEREDMODEL = DESCRIPTOR.message_types_by_name['CreateRegisteredModel']
_CREATEREGISTEREDMODEL_RESPONSE = _CREATEREGISTEREDMODEL.nested_types_by_name['Response']
_RENAMEREGISTEREDMODEL = DESCRIPTOR.message_types_by_name['RenameRegisteredModel']
_RENAMEREGISTEREDMODEL_RESPONSE = _RENAMEREGISTEREDMODEL.nested_types_by_name['Response']
_UPDATEREGISTEREDMODEL = DESCRIPTOR.message_types_by_name['UpdateRegisteredModel']
_UPDATEREGISTEREDMODEL_RESPONSE = _UPDATEREGISTEREDMODEL.nested_types_by_name['Response']
_DELETEREGISTEREDMODEL = DESCRIPTOR.message_types_by_name['DeleteRegisteredModel']
_DELETEREGISTEREDMODEL_RESPONSE = _DELETEREGISTEREDMODEL.nested_types_by_name['Response']
_GETREGISTEREDMODEL = DESCRIPTOR.message_types_by_name['GetRegisteredModel']
_GETREGISTEREDMODEL_RESPONSE = _GETREGISTEREDMODEL.nested_types_by_name['Response']
_SEARCHREGISTEREDMODELS = DESCRIPTOR.message_types_by_name['SearchRegisteredModels']
_SEARCHREGISTEREDMODELS_RESPONSE = _SEARCHREGISTEREDMODELS.nested_types_by_name['Response']
_GETLATESTVERSIONS = DESCRIPTOR.message_types_by_name['GetLatestVersions']
_GETLATESTVERSIONS_RESPONSE = _GETLATESTVERSIONS.nested_types_by_name['Response']
_CREATEMODELVERSION = DESCRIPTOR.message_types_by_name['CreateModelVersion']
_CREATEMODELVERSION_RESPONSE = _CREATEMODELVERSION.nested_types_by_name['Response']
_UPDATEMODELVERSION = DESCRIPTOR.message_types_by_name['UpdateModelVersion']
_UPDATEMODELVERSION_RESPONSE = _UPDATEMODELVERSION.nested_types_by_name['Response']
_TRANSITIONMODELVERSIONSTAGE = DESCRIPTOR.message_types_by_name['TransitionModelVersionStage']
_TRANSITIONMODELVERSIONSTAGE_RESPONSE = _TRANSITIONMODELVERSIONSTAGE.nested_types_by_name['Response']
_DELETEMODELVERSION = DESCRIPTOR.message_types_by_name['DeleteModelVersion']
_DELETEMODELVERSION_RESPONSE = _DELETEMODELVERSION.nested_types_by_name['Response']
_GETMODELVERSION = DESCRIPTOR.message_types_by_name['GetModelVersion']
_GETMODELVERSION_RESPONSE = _GETMODELVERSION.nested_types_by_name['Response']
_SEARCHMODELVERSIONS = DESCRIPTOR.message_types_by_name['SearchModelVersions']
_SEARCHMODELVERSIONS_RESPONSE = _SEARCHMODELVERSIONS.nested_types_by_name['Response']
_GETMODELVERSIONDOWNLOADURI = DESCRIPTOR.message_types_by_name['GetModelVersionDownloadUri']
_GETMODELVERSIONDOWNLOADURI_RESPONSE = _GETMODELVERSIONDOWNLOADURI.nested_types_by_name['Response']
_MODELVERSIONTAG = DESCRIPTOR.message_types_by_name['ModelVersionTag']
_REGISTEREDMODELTAG = DESCRIPTOR.message_types_by_name['RegisteredModelTag']
_SETREGISTEREDMODELTAG = DESCRIPTOR.message_types_by_name['SetRegisteredModelTag']
_SETREGISTEREDMODELTAG_RESPONSE = _SETREGISTEREDMODELTAG.nested_types_by_name['Response']
_SETMODELVERSIONTAG = DESCRIPTOR.message_types_by_name['SetModelVersionTag']
_SETMODELVERSIONTAG_RESPONSE = _SETMODELVERSIONTAG.nested_types_by_name['Response']
_DELETEREGISTEREDMODELTAG = DESCRIPTOR.message_types_by_name['DeleteRegisteredModelTag']
_DELETEREGISTEREDMODELTAG_RESPONSE = _DELETEREGISTEREDMODELTAG.nested_types_by_name['Response']
_DELETEMODELVERSIONTAG = DESCRIPTOR.message_types_by_name['DeleteModelVersionTag']
_DELETEMODELVERSIONTAG_RESPONSE = _DELETEMODELVERSIONTAG.nested_types_by_name['Response']
_REGISTEREDMODELALIAS = DESCRIPTOR.message_types_by_name['RegisteredModelAlias']
_SETREGISTEREDMODELALIAS = DESCRIPTOR.message_types_by_name['SetRegisteredModelAlias']
_SETREGISTEREDMODELALIAS_RESPONSE = _SETREGISTEREDMODELALIAS.nested_types_by_name['Response']
_DELETEREGISTEREDMODELALIAS = DESCRIPTOR.message_types_by_name['DeleteRegisteredModelAlias']
_DELETEREGISTEREDMODELALIAS_RESPONSE = _DELETEREGISTEREDMODELALIAS.nested_types_by_name['Response']
_GETMODELVERSIONBYALIAS = DESCRIPTOR.message_types_by_name['GetModelVersionByAlias']
_GETMODELVERSIONBYALIAS_RESPONSE = _GETMODELVERSIONBYALIAS.nested_types_by_name['Response']
RegisteredModel = _reflection.GeneratedProtocolMessageType('RegisteredModel', (_message.Message,), {
  'DESCRIPTOR' : _REGISTEREDMODEL,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RegisteredModel)
  })
_sym_db.RegisterMessage(RegisteredModel)

ModelVersion = _reflection.GeneratedProtocolMessageType('ModelVersion', (_message.Message,), {
  'DESCRIPTOR' : _MODELVERSION,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.ModelVersion)
  })
_sym_db.RegisterMessage(ModelVersion)

CreateRegisteredModel = _reflection.GeneratedProtocolMessageType('CreateRegisteredModel', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _CREATEREGISTEREDMODEL_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.CreateRegisteredModel.Response)
    })
  ,
  'DESCRIPTOR' : _CREATEREGISTEREDMODEL,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.CreateRegisteredModel)
  })
_sym_db.RegisterMessage(CreateRegisteredModel)
_sym_db.RegisterMessage(CreateRegisteredModel.Response)

RenameRegisteredModel = _reflection.GeneratedProtocolMessageType('RenameRegisteredModel', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _RENAMEREGISTEREDMODEL_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.RenameRegisteredModel.Response)
    })
  ,
  'DESCRIPTOR' : _RENAMEREGISTEREDMODEL,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RenameRegisteredModel)
  })
_sym_db.RegisterMessage(RenameRegisteredModel)
_sym_db.RegisterMessage(RenameRegisteredModel.Response)

UpdateRegisteredModel = _reflection.GeneratedProtocolMessageType('UpdateRegisteredModel', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEREGISTEREDMODEL_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.UpdateRegisteredModel.Response)
    })
  ,
  'DESCRIPTOR' : _UPDATEREGISTEREDMODEL,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.UpdateRegisteredModel)
  })
_sym_db.RegisterMessage(UpdateRegisteredModel)
_sym_db.RegisterMessage(UpdateRegisteredModel.Response)

DeleteRegisteredModel = _reflection.GeneratedProtocolMessageType('DeleteRegisteredModel', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETEREGISTEREDMODEL_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteRegisteredModel.Response)
    })
  ,
  'DESCRIPTOR' : _DELETEREGISTEREDMODEL,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteRegisteredModel)
  })
_sym_db.RegisterMessage(DeleteRegisteredModel)
_sym_db.RegisterMessage(DeleteRegisteredModel.Response)

GetRegisteredModel = _reflection.GeneratedProtocolMessageType('GetRegisteredModel', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETREGISTEREDMODEL_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetRegisteredModel.Response)
    })
  ,
  'DESCRIPTOR' : _GETREGISTEREDMODEL,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetRegisteredModel)
  })
_sym_db.RegisterMessage(GetRegisteredModel)
_sym_db.RegisterMessage(GetRegisteredModel.Response)

SearchRegisteredModels = _reflection.GeneratedProtocolMessageType('SearchRegisteredModels', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHREGISTEREDMODELS_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SearchRegisteredModels.Response)
    })
  ,
  'DESCRIPTOR' : _SEARCHREGISTEREDMODELS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SearchRegisteredModels)
  })
_sym_db.RegisterMessage(SearchRegisteredModels)
_sym_db.RegisterMessage(SearchRegisteredModels.Response)

GetLatestVersions = _reflection.GeneratedProtocolMessageType('GetLatestVersions', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETLATESTVERSIONS_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetLatestVersions.Response)
    })
  ,
  'DESCRIPTOR' : _GETLATESTVERSIONS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetLatestVersions)
  })
_sym_db.RegisterMessage(GetLatestVersions)
_sym_db.RegisterMessage(GetLatestVersions.Response)

CreateModelVersion = _reflection.GeneratedProtocolMessageType('CreateModelVersion', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _CREATEMODELVERSION_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.CreateModelVersion.Response)
    })
  ,
  'DESCRIPTOR' : _CREATEMODELVERSION,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.CreateModelVersion)
  })
_sym_db.RegisterMessage(CreateModelVersion)
_sym_db.RegisterMessage(CreateModelVersion.Response)

UpdateModelVersion = _reflection.GeneratedProtocolMessageType('UpdateModelVersion', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEMODELVERSION_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.UpdateModelVersion.Response)
    })
  ,
  'DESCRIPTOR' : _UPDATEMODELVERSION,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.UpdateModelVersion)
  })
_sym_db.RegisterMessage(UpdateModelVersion)
_sym_db.RegisterMessage(UpdateModelVersion.Response)

TransitionModelVersionStage = _reflection.GeneratedProtocolMessageType('TransitionModelVersionStage', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _TRANSITIONMODELVERSIONSTAGE_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.TransitionModelVersionStage.Response)
    })
  ,
  'DESCRIPTOR' : _TRANSITIONMODELVERSIONSTAGE,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.TransitionModelVersionStage)
  })
_sym_db.RegisterMessage(TransitionModelVersionStage)
_sym_db.RegisterMessage(TransitionModelVersionStage.Response)

DeleteModelVersion = _reflection.GeneratedProtocolMessageType('DeleteModelVersion', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETEMODELVERSION_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteModelVersion.Response)
    })
  ,
  'DESCRIPTOR' : _DELETEMODELVERSION,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteModelVersion)
  })
_sym_db.RegisterMessage(DeleteModelVersion)
_sym_db.RegisterMessage(DeleteModelVersion.Response)

GetModelVersion = _reflection.GeneratedProtocolMessageType('GetModelVersion', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETMODELVERSION_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetModelVersion.Response)
    })
  ,
  'DESCRIPTOR' : _GETMODELVERSION,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetModelVersion)
  })
_sym_db.RegisterMessage(GetModelVersion)
_sym_db.RegisterMessage(GetModelVersion.Response)

SearchModelVersions = _reflection.GeneratedProtocolMessageType('SearchModelVersions', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHMODELVERSIONS_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SearchModelVersions.Response)
    })
  ,
  'DESCRIPTOR' : _SEARCHMODELVERSIONS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SearchModelVersions)
  })
_sym_db.RegisterMessage(SearchModelVersions)
_sym_db.RegisterMessage(SearchModelVersions.Response)

GetModelVersionDownloadUri = _reflection.GeneratedProtocolMessageType('GetModelVersionDownloadUri', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETMODELVERSIONDOWNLOADURI_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetModelVersionDownloadUri.Response)
    })
  ,
  'DESCRIPTOR' : _GETMODELVERSIONDOWNLOADURI,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetModelVersionDownloadUri)
  })
_sym_db.RegisterMessage(GetModelVersionDownloadUri)
_sym_db.RegisterMessage(GetModelVersionDownloadUri.Response)

ModelVersionTag = _reflection.GeneratedProtocolMessageType('ModelVersionTag', (_message.Message,), {
  'DESCRIPTOR' : _MODELVERSIONTAG,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.ModelVersionTag)
  })
_sym_db.RegisterMessage(ModelVersionTag)

RegisteredModelTag = _reflection.GeneratedProtocolMessageType('RegisteredModelTag', (_message.Message,), {
  'DESCRIPTOR' : _REGISTEREDMODELTAG,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RegisteredModelTag)
  })
_sym_db.RegisterMessage(RegisteredModelTag)

SetRegisteredModelTag = _reflection.GeneratedProtocolMessageType('SetRegisteredModelTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SETREGISTEREDMODELTAG_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SetRegisteredModelTag.Response)
    })
  ,
  'DESCRIPTOR' : _SETREGISTEREDMODELTAG,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SetRegisteredModelTag)
  })
_sym_db.RegisterMessage(SetRegisteredModelTag)
_sym_db.RegisterMessage(SetRegisteredModelTag.Response)

SetModelVersionTag = _reflection.GeneratedProtocolMessageType('SetModelVersionTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SETMODELVERSIONTAG_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SetModelVersionTag.Response)
    })
  ,
  'DESCRIPTOR' : _SETMODELVERSIONTAG,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SetModelVersionTag)
  })
_sym_db.RegisterMessage(SetModelVersionTag)
_sym_db.RegisterMessage(SetModelVersionTag.Response)

DeleteRegisteredModelTag = _reflection.GeneratedProtocolMessageType('DeleteRegisteredModelTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETEREGISTEREDMODELTAG_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteRegisteredModelTag.Response)
    })
  ,
  'DESCRIPTOR' : _DELETEREGISTEREDMODELTAG,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteRegisteredModelTag)
  })
_sym_db.RegisterMessage(DeleteRegisteredModelTag)
_sym_db.RegisterMessage(DeleteRegisteredModelTag.Response)

DeleteModelVersionTag = _reflection.GeneratedProtocolMessageType('DeleteModelVersionTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETEMODELVERSIONTAG_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteModelVersionTag.Response)
    })
  ,
  'DESCRIPTOR' : _DELETEMODELVERSIONTAG,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteModelVersionTag)
  })
_sym_db.RegisterMessage(DeleteModelVersionTag)
_sym_db.RegisterMessage(DeleteModelVersionTag.Response)

RegisteredModelAlias = _reflection.GeneratedProtocolMessageType('RegisteredModelAlias', (_message.Message,), {
  'DESCRIPTOR' : _REGISTEREDMODELALIAS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RegisteredModelAlias)
  })
_sym_db.RegisterMessage(RegisteredModelAlias)

SetRegisteredModelAlias = _reflection.GeneratedProtocolMessageType('SetRegisteredModelAlias', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SETREGISTEREDMODELALIAS_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SetRegisteredModelAlias.Response)
    })
  ,
  'DESCRIPTOR' : _SETREGISTEREDMODELALIAS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SetRegisteredModelAlias)
  })
_sym_db.RegisterMessage(SetRegisteredModelAlias)
_sym_db.RegisterMessage(SetRegisteredModelAlias.Response)

DeleteRegisteredModelAlias = _reflection.GeneratedProtocolMessageType('DeleteRegisteredModelAlias', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETEREGISTEREDMODELALIAS_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteRegisteredModelAlias.Response)
    })
  ,
  'DESCRIPTOR' : _DELETEREGISTEREDMODELALIAS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteRegisteredModelAlias)
  })
_sym_db.RegisterMessage(DeleteRegisteredModelAlias)
_sym_db.RegisterMessage(DeleteRegisteredModelAlias.Response)

GetModelVersionByAlias = _reflection.GeneratedProtocolMessageType('GetModelVersionByAlias', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETMODELVERSIONBYALIAS_RESPONSE,
    '__module__' : 'model_registry_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetModelVersionByAlias.Response)
    })
  ,
  'DESCRIPTOR' : _GETMODELVERSIONBYALIAS,
  '__module__' : 'model_registry_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetModelVersionByAlias)
  })
_sym_db.RegisterMessage(GetModelVersionByAlias)
_sym_db.RegisterMessage(GetModelVersionByAlias.Response)

_MODELREGISTRYSERVICE = DESCRIPTOR.services_by_name['ModelRegistryService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024org.mlflow.api.proto\220\001\001\240\001\001\342?\002\020\001'
  _CREATEREGISTEREDMODEL.fields_by_name['name']._options = None
  _CREATEREGISTEREDMODEL.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _CREATEREGISTEREDMODEL._options = None
  _CREATEREGISTEREDMODEL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _RENAMEREGISTEREDMODEL.fields_by_name['name']._options = None
  _RENAMEREGISTEREDMODEL.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _RENAMEREGISTEREDMODEL._options = None
  _RENAMEREGISTEREDMODEL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _UPDATEREGISTEREDMODEL.fields_by_name['name']._options = None
  _UPDATEREGISTEREDMODEL.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _UPDATEREGISTEREDMODEL._options = None
  _UPDATEREGISTEREDMODEL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETEREGISTEREDMODEL.fields_by_name['name']._options = None
  _DELETEREGISTEREDMODEL.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _DELETEREGISTEREDMODEL._options = None
  _DELETEREGISTEREDMODEL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETREGISTEREDMODEL.fields_by_name['name']._options = None
  _GETREGISTEREDMODEL.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _GETREGISTEREDMODEL._options = None
  _GETREGISTEREDMODEL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SEARCHREGISTEREDMODELS._options = None
  _SEARCHREGISTEREDMODELS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETLATESTVERSIONS.fields_by_name['name']._options = None
  _GETLATESTVERSIONS.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _GETLATESTVERSIONS._options = None
  _GETLATESTVERSIONS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _CREATEMODELVERSION.fields_by_name['name']._options = None
  _CREATEMODELVERSION.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _CREATEMODELVERSION.fields_by_name['source']._options = None
  _CREATEMODELVERSION.fields_by_name['source']._serialized_options = b'\370\206\031\001'
  _CREATEMODELVERSION._options = None
  _CREATEMODELVERSION._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _UPDATEMODELVERSION.fields_by_name['name']._options = None
  _UPDATEMODELVERSION.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _UPDATEMODELVERSION.fields_by_name['version']._options = None
  _UPDATEMODELVERSION.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _UPDATEMODELVERSION._options = None
  _UPDATEMODELVERSION._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['name']._options = None
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['version']._options = None
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['stage']._options = None
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['stage']._serialized_options = b'\370\206\031\001'
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['archive_existing_versions']._options = None
  _TRANSITIONMODELVERSIONSTAGE.fields_by_name['archive_existing_versions']._serialized_options = b'\370\206\031\001'
  _TRANSITIONMODELVERSIONSTAGE._options = None
  _TRANSITIONMODELVERSIONSTAGE._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETEMODELVERSION.fields_by_name['name']._options = None
  _DELETEMODELVERSION.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _DELETEMODELVERSION.fields_by_name['version']._options = None
  _DELETEMODELVERSION.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _DELETEMODELVERSION._options = None
  _DELETEMODELVERSION._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETMODELVERSION.fields_by_name['name']._options = None
  _GETMODELVERSION.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _GETMODELVERSION.fields_by_name['version']._options = None
  _GETMODELVERSION.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _GETMODELVERSION._options = None
  _GETMODELVERSION._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SEARCHMODELVERSIONS._options = None
  _SEARCHMODELVERSIONS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETMODELVERSIONDOWNLOADURI.fields_by_name['name']._options = None
  _GETMODELVERSIONDOWNLOADURI.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _GETMODELVERSIONDOWNLOADURI.fields_by_name['version']._options = None
  _GETMODELVERSIONDOWNLOADURI.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _GETMODELVERSIONDOWNLOADURI._options = None
  _GETMODELVERSIONDOWNLOADURI._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SETREGISTEREDMODELTAG.fields_by_name['name']._options = None
  _SETREGISTEREDMODELTAG.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _SETREGISTEREDMODELTAG.fields_by_name['key']._options = None
  _SETREGISTEREDMODELTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _SETREGISTEREDMODELTAG.fields_by_name['value']._options = None
  _SETREGISTEREDMODELTAG.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _SETREGISTEREDMODELTAG._options = None
  _SETREGISTEREDMODELTAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SETMODELVERSIONTAG.fields_by_name['name']._options = None
  _SETMODELVERSIONTAG.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _SETMODELVERSIONTAG.fields_by_name['version']._options = None
  _SETMODELVERSIONTAG.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _SETMODELVERSIONTAG.fields_by_name['key']._options = None
  _SETMODELVERSIONTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _SETMODELVERSIONTAG.fields_by_name['value']._options = None
  _SETMODELVERSIONTAG.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _SETMODELVERSIONTAG._options = None
  _SETMODELVERSIONTAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETEREGISTEREDMODELTAG.fields_by_name['name']._options = None
  _DELETEREGISTEREDMODELTAG.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _DELETEREGISTEREDMODELTAG.fields_by_name['key']._options = None
  _DELETEREGISTEREDMODELTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _DELETEREGISTEREDMODELTAG._options = None
  _DELETEREGISTEREDMODELTAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETEMODELVERSIONTAG.fields_by_name['name']._options = None
  _DELETEMODELVERSIONTAG.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _DELETEMODELVERSIONTAG.fields_by_name['version']._options = None
  _DELETEMODELVERSIONTAG.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _DELETEMODELVERSIONTAG.fields_by_name['key']._options = None
  _DELETEMODELVERSIONTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _DELETEMODELVERSIONTAG._options = None
  _DELETEMODELVERSIONTAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SETREGISTEREDMODELALIAS.fields_by_name['name']._options = None
  _SETREGISTEREDMODELALIAS.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _SETREGISTEREDMODELALIAS.fields_by_name['alias']._options = None
  _SETREGISTEREDMODELALIAS.fields_by_name['alias']._serialized_options = b'\370\206\031\001'
  _SETREGISTEREDMODELALIAS.fields_by_name['version']._options = None
  _SETREGISTEREDMODELALIAS.fields_by_name['version']._serialized_options = b'\370\206\031\001'
  _SETREGISTEREDMODELALIAS._options = None
  _SETREGISTEREDMODELALIAS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETEREGISTEREDMODELALIAS.fields_by_name['name']._options = None
  _DELETEREGISTEREDMODELALIAS.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _DELETEREGISTEREDMODELALIAS.fields_by_name['alias']._options = None
  _DELETEREGISTEREDMODELALIAS.fields_by_name['alias']._serialized_options = b'\370\206\031\001'
  _DELETEREGISTEREDMODELALIAS._options = None
  _DELETEREGISTEREDMODELALIAS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETMODELVERSIONBYALIAS.fields_by_name['name']._options = None
  _GETMODELVERSIONBYALIAS.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _GETMODELVERSIONBYALIAS.fields_by_name['alias']._options = None
  _GETMODELVERSIONBYALIAS.fields_by_name['alias']._serialized_options = b'\370\206\031\001'
  _GETMODELVERSIONBYALIAS._options = None
  _GETMODELVERSIONBYALIAS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _MODELREGISTRYSERVICE.methods_by_name['createRegisteredModel']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['createRegisteredModel']._serialized_options = b'\362\206\031J\n.\n\004POST\022 /mlflow/registered-models/create\032\004\010\002\020\000\020\001*\026Create RegisteredModel'
  _MODELREGISTRYSERVICE.methods_by_name['renameRegisteredModel']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['renameRegisteredModel']._serialized_options = b'\362\206\031J\n.\n\004POST\022 /mlflow/registered-models/rename\032\004\010\002\020\000\020\001*\026Rename RegisteredModel'
  _MODELREGISTRYSERVICE.methods_by_name['updateRegisteredModel']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['updateRegisteredModel']._serialized_options = b'\362\206\031K\n/\n\005PATCH\022 /mlflow/registered-models/update\032\004\010\002\020\000\020\001*\026Update RegisteredModel'
  _MODELREGISTRYSERVICE.methods_by_name['deleteRegisteredModel']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['deleteRegisteredModel']._serialized_options = b'\362\206\031L\n0\n\006DELETE\022 /mlflow/registered-models/delete\032\004\010\002\020\000\020\001*\026Delete RegisteredModel'
  _MODELREGISTRYSERVICE.methods_by_name['getRegisteredModel']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['getRegisteredModel']._serialized_options = b'\362\206\031C\n*\n\003GET\022\035/mlflow/registered-models/get\032\004\010\002\020\000\020\001*\023Get RegisteredModel'
  _MODELREGISTRYSERVICE.methods_by_name['searchRegisteredModels']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['searchRegisteredModels']._serialized_options = b'\362\206\031J\n-\n\003GET\022 /mlflow/registered-models/search\032\004\010\002\020\000\020\001*\027Search RegisteredModels'
  _MODELREGISTRYSERVICE.methods_by_name['getLatestVersions']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['getLatestVersions']._serialized_options = b'\362\206\031\225\001\n;\n\004POST\022-/mlflow/registered-models/get-latest-versions\032\004\010\002\020\000\n:\n\003GET\022-/mlflow/registered-models/get-latest-versions\032\004\010\002\020\000\020\001*\030Get Latest ModelVersions'
  _MODELREGISTRYSERVICE.methods_by_name['createModelVersion']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['createModelVersion']._serialized_options = b'\362\206\031D\n+\n\004POST\022\035/mlflow/model-versions/create\032\004\010\002\020\000\020\001*\023Create ModelVersion'
  _MODELREGISTRYSERVICE.methods_by_name['updateModelVersion']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['updateModelVersion']._serialized_options = b'\362\206\031E\n,\n\005PATCH\022\035/mlflow/model-versions/update\032\004\010\002\020\000\020\001*\023Update ModelVersion'
  _MODELREGISTRYSERVICE.methods_by_name['transitionModelVersionStage']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['transitionModelVersionStage']._serialized_options = b'\362\206\031X\n5\n\004POST\022\'/mlflow/model-versions/transition-stage\032\004\010\002\020\000\020\001*\035Transition ModelVersion Stage'
  _MODELREGISTRYSERVICE.methods_by_name['deleteModelVersion']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['deleteModelVersion']._serialized_options = b'\362\206\031F\n-\n\006DELETE\022\035/mlflow/model-versions/delete\032\004\010\002\020\000\020\001*\023Delete ModelVersion'
  _MODELREGISTRYSERVICE.methods_by_name['getModelVersion']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['getModelVersion']._serialized_options = b'\362\206\031=\n\'\n\003GET\022\032/mlflow/model-versions/get\032\004\010\002\020\000\020\001*\020Get ModelVersion'
  _MODELREGISTRYSERVICE.methods_by_name['searchModelVersions']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['searchModelVersions']._serialized_options = b'\362\206\031D\n*\n\003GET\022\035/mlflow/model-versions/search\032\004\010\002\020\000\020\001*\024Search ModelVersions\272\214\031\000'
  _MODELREGISTRYSERVICE.methods_by_name['getModelVersionDownloadUri']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['getModelVersionDownloadUri']._serialized_options = b'\362\206\031e\n4\n\003GET\022\'/mlflow/model-versions/get-download-uri\032\004\010\002\020\000\020\001*+Get Download URI For ModelVersion Artifacts'
  _MODELREGISTRYSERVICE.methods_by_name['setRegisteredModelTag']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['setRegisteredModelTag']._serialized_options = b'\362\206\031M\n/\n\004POST\022!/mlflow/registered-models/set-tag\032\004\010\002\020\000\020\001*\030Set Registered Model Tag'
  _MODELREGISTRYSERVICE.methods_by_name['setModelVersionTag']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['setModelVersionTag']._serialized_options = b'\362\206\031G\n,\n\004POST\022\036/mlflow/model-versions/set-tag\032\004\010\002\020\000\020\001*\025Set Model Version Tag'
  _MODELREGISTRYSERVICE.methods_by_name['deleteRegisteredModelTag']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['deleteRegisteredModelTag']._serialized_options = b'\362\206\031U\n4\n\006DELETE\022$/mlflow/registered-models/delete-tag\032\004\010\002\020\000\020\001*\033Delete Registered Model Tag'
  _MODELREGISTRYSERVICE.methods_by_name['deleteModelVersionTag']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['deleteModelVersionTag']._serialized_options = b'\362\206\031O\n1\n\006DELETE\022!/mlflow/model-versions/delete-tag\032\004\010\002\020\000\020\001*\030Delete Model Version Tag'
  _MODELREGISTRYSERVICE.methods_by_name['setRegisteredModelAlias']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['setRegisteredModelAlias']._serialized_options = b'\362\206\031M\n-\n\004POST\022\037/mlflow/registered-models/alias\032\004\010\002\020\000\020\001*\032Set Registered Model Alias'
  _MODELREGISTRYSERVICE.methods_by_name['deleteRegisteredModelAlias']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['deleteRegisteredModelAlias']._serialized_options = b'\362\206\031R\n/\n\006DELETE\022\037/mlflow/registered-models/alias\032\004\010\002\020\000\020\001*\035Delete Registered Model Alias'
  _MODELREGISTRYSERVICE.methods_by_name['getModelVersionByAlias']._options = None
  _MODELREGISTRYSERVICE.methods_by_name['getModelVersionByAlias']._serialized_options = b'\362\206\031L\n,\n\003GET\022\037/mlflow/registered-models/alias\032\004\010\002\020\000\020\001*\032Get Model Version by Alias'
  _MODELVERSIONSTATUS._serialized_start=4442
  _MODELVERSIONSTATUS._serialized_end=4524
  _REGISTEREDMODEL._serialized_start=74
  _REGISTEREDMODEL._serialized_end=339
  _MODELVERSION._serialized_start=342
  _MODELVERSION._serialized_end=682
  _CREATEREGISTEREDMODEL._serialized_start=685
  _CREATEREGISTEREDMODEL._serialized_end=899
  _CREATEREGISTEREDMODEL_RESPONSE._serialized_start=793
  _CREATEREGISTEREDMODEL_RESPONSE._serialized_end=854
  _RENAMEREGISTEREDMODEL._serialized_start=902
  _RENAMEREGISTEREDMODEL._serialized_end=1071
  _RENAMEREGISTEREDMODEL_RESPONSE._serialized_start=793
  _RENAMEREGISTEREDMODEL_RESPONSE._serialized_end=854
  _UPDATEREGISTEREDMODEL._serialized_start=1074
  _UPDATEREGISTEREDMODEL._serialized_end=1246
  _UPDATEREGISTEREDMODEL_RESPONSE._serialized_start=793
  _UPDATEREGISTEREDMODEL_RESPONSE._serialized_end=854
  _DELETEREGISTEREDMODEL._serialized_start=1248
  _DELETEREGISTEREDMODEL._serialized_end=1348
  _DELETEREGISTEREDMODEL_RESPONSE._serialized_start=793
  _DELETEREGISTEREDMODEL_RESPONSE._serialized_end=803
  _GETREGISTEREDMODEL._serialized_start=1351
  _GETREGISTEREDMODEL._serialized_end=1499
  _GETREGISTEREDMODEL_RESPONSE._serialized_start=793
  _GETREGISTEREDMODEL_RESPONSE._serialized_end=854
  _SEARCHREGISTEREDMODELS._serialized_start=1502
  _SEARCHREGISTEREDMODELS._serialized_end=1740
  _SEARCHREGISTEREDMODELS_RESPONSE._serialized_start=1608
  _SEARCHREGISTEREDMODELS_RESPONSE._serialized_end=1695
  _GETLATESTVERSIONS._serialized_start=1743
  _GETLATESTVERSIONS._serialized_end=1901
  _GETLATESTVERSIONS_RESPONSE._serialized_start=1800
  _GETLATESTVERSIONS_RESPONSE._serialized_end=1856
  _CREATEMODELVERSION._serialized_start=1904
  _CREATEMODELVERSION._serialized_end=2162
  _CREATEMODELVERSION_RESPONSE._serialized_start=2062
  _CREATEMODELVERSION_RESPONSE._serialized_end=2117
  _UPDATEMODELVERSION._serialized_start=2165
  _UPDATEMODELVERSION._serialized_end=2351
  _UPDATEMODELVERSION_RESPONSE._serialized_start=2062
  _UPDATEMODELVERSION_RESPONSE._serialized_end=2117
  _TRANSITIONMODELVERSIONSTAGE._serialized_start=2354
  _TRANSITIONMODELVERSIONSTAGE._serialized_end=2590
  _TRANSITIONMODELVERSIONSTAGE_RESPONSE._serialized_start=2062
  _TRANSITIONMODELVERSIONSTAGE_RESPONSE._serialized_end=2117
  _DELETEMODELVERSION._serialized_start=2592
  _DELETEMODELVERSION._serialized_end=2712
  _DELETEMODELVERSION_RESPONSE._serialized_start=793
  _DELETEMODELVERSION_RESPONSE._serialized_end=803
  _GETMODELVERSION._serialized_start=2715
  _GETMODELVERSION._serialized_end=2877
  _GETMODELVERSION_RESPONSE._serialized_start=2062
  _GETMODELVERSION_RESPONSE._serialized_end=2117
  _SEARCHMODELVERSIONS._serialized_start=2880
  _SEARCHMODELVERSIONS._serialized_end=3112
  _SEARCHMODELVERSIONS_RESPONSE._serialized_start=2986
  _SEARCHMODELVERSIONS_RESPONSE._serialized_end=3067
  _GETMODELVERSIONDOWNLOADURI._serialized_start=3115
  _GETMODELVERSIONDOWNLOADURI._serialized_end=3265
  _GETMODELVERSIONDOWNLOADURI_RESPONSE._serialized_start=3188
  _GETMODELVERSIONDOWNLOADURI_RESPONSE._serialized_end=3220
  _MODELVERSIONTAG._serialized_start=3267
  _MODELVERSIONTAG._serialized_end=3312
  _REGISTEREDMODELTAG._serialized_start=3314
  _REGISTEREDMODELTAG._serialized_end=3362
  _SETREGISTEREDMODELTAG._serialized_start=3365
  _SETREGISTEREDMODELTAG._serialized_end=3505
  _SETREGISTEREDMODELTAG_RESPONSE._serialized_start=793
  _SETREGISTEREDMODELTAG_RESPONSE._serialized_end=803
  _SETMODELVERSIONTAG._serialized_start=3508
  _SETMODELVERSIONTAG._serialized_end=3668
  _SETMODELVERSIONTAG_RESPONSE._serialized_start=793
  _SETMODELVERSIONTAG_RESPONSE._serialized_end=803
  _DELETEREGISTEREDMODELTAG._serialized_start=3670
  _DELETEREGISTEREDMODELTAG._serialized_end=3792
  _DELETEREGISTEREDMODELTAG_RESPONSE._serialized_start=793
  _DELETEREGISTEREDMODELTAG_RESPONSE._serialized_end=803
  _DELETEMODELVERSIONTAG._serialized_start=3795
  _DELETEMODELVERSIONTAG._serialized_end=3937
  _DELETEMODELVERSIONTAG_RESPONSE._serialized_start=793
  _DELETEMODELVERSIONTAG_RESPONSE._serialized_end=803
  _REGISTEREDMODELALIAS._serialized_start=3939
  _REGISTEREDMODELALIAS._serialized_end=3993
  _SETREGISTEREDMODELALIAS._serialized_start=3996
  _SETREGISTEREDMODELALIAS._serialized_end=4142
  _SETREGISTEREDMODELALIAS_RESPONSE._serialized_start=793
  _SETREGISTEREDMODELALIAS_RESPONSE._serialized_end=803
  _DELETEREGISTEREDMODELALIAS._serialized_start=4144
  _DELETEREGISTEREDMODELALIAS._serialized_end=4270
  _DELETEREGISTEREDMODELALIAS_RESPONSE._serialized_start=793
  _DELETEREGISTEREDMODELALIAS_RESPONSE._serialized_end=803
  _GETMODELVERSIONBYALIAS._serialized_start=4273
  _GETMODELVERSIONBYALIAS._serialized_end=4440
  _GETMODELVERSIONBYALIAS_RESPONSE._serialized_start=2062
  _GETMODELVERSIONBYALIAS_RESPONSE._serialized_end=2117
  _MODELREGISTRYSERVICE._serialized_start=4527
  _MODELREGISTRYSERVICE._serialized_end=8367
ModelRegistryService = service_reflection.GeneratedServiceType('ModelRegistryService', (_service.Service,), dict(
  DESCRIPTOR = _MODELREGISTRYSERVICE,
  __module__ = 'model_registry_pb2'
  ))

ModelRegistryService_Stub = service_reflection.GeneratedServiceStubType('ModelRegistryService_Stub', (ModelRegistryService,), dict(
  DESCRIPTOR = _MODELREGISTRYSERVICE,
  __module__ = 'model_registry_pb2'
  ))


# @@protoc_insertion_point(module_scope)



import importlib
import logging
import os
import sys
import time

import cloudpickle
from packaging.version import Version

from mlflow.exceptions import INVALID_PARAMETER_VALUE, MlflowException
from mlflow.recipes.artifacts import DataframeArtifact, TransformerArtifact
from mlflow.recipes.cards import BaseCard
from mlflow.recipes.step import BaseStep, StepClass
from mlflow.recipes.utils.execution import get_step_output_path
from mlflow.recipes.utils.step import get_pandas_data_profiles, validate_classification_config
from mlflow.recipes.utils.tracking import TrackingConfig, get_recipe_tracking_config

_logger = logging.getLogger(__name__)

_USER_DEFINED_TRANSFORM_STEP_MODULE = "steps.transform"


def _generate_feature_names(num_features):
    max_length = len(str(num_features))
    return ["f_" + str(i).zfill(max_length) for i in range(num_features)]


def _get_output_feature_names(transformer, num_features, input_features):
    import sklearn

    # `get_feature_names_out` was introduced in scikit-learn 1.0.0.
    if Version(sklearn.__version__) < Version("1.0.0"):
        return _generate_feature_names(num_features)

    try:
        # `get_feature_names_out` fails if `transformer` contains a transformer that doesn't
        # implement `get_feature_names_out`. For example, `FunctionTransformer` only implements
        # `get_feature_names_out` when it's instantiated with `feature_names_out`.
        # In scikit-learn >= 1.1.0, all transformers implement `get_feature_names_out`.
        # In scikit-learn == 1.0.*, some transformers implement `get_feature_names_out`.
        return transformer.get_feature_names_out(input_features)
    except Exception as e:
        _logger.warning(
            f"Failed to get output feature names with `get_feature_names_out`: {e}. "
            "Falling back to using auto-generated feature names."
        )
        return _generate_feature_names(num_features)


def _validate_user_code_output(transformer_fn):
    transformer = transformer_fn()
    if transformer is not None and not (hasattr(transformer, "fit") and callable(transformer.fit)):
        raise MlflowException(
            message="The transformer provided doesn't have a fit method."
        ) from None

    if transformer is not None and not (
        hasattr(transformer, "transform") and callable(transformer.transform)
    ):
        raise MlflowException(
            message="The transformer provided doesn't have a transform method."
        ) from None

    return transformer


class TransformStep(BaseStep):
    def __init__(self, step_config, recipe_root):
        super().__init__(step_config, recipe_root)
        self.tracking_config = TrackingConfig.from_dict(self.step_config)

    def _validate_and_apply_step_config(self):
        self.target_col = self.step_config.get("target_col")
        self.positive_class = self.step_config.get("positive_class")
        if self.target_col is None:
            raise MlflowException(
                "Missing target_col config in recipe config.",
                error_code=INVALID_PARAMETER_VALUE,
            )
        if "using" in self.step_config:
            if self.step_config["using"] not in ["custom"]:
                raise MlflowException(
                    f"Invalid transform step configuration value {self.step_config['using']} for "
                    f"key 'using'. Supported values are: ['custom']",
                    error_code=INVALID_PARAMETER_VALUE,
                )
        else:
            self.step_config["using"] = "custom"
        self.run_end_time = None
        self.execution_duration = None
        self.skip_data_profiling = self.step_config.get("skip_data_profiling", False)

    def _run(self, output_directory):
        import pandas as pd

        run_start_time = time.time()

        train_data_path = get_step_output_path(
            recipe_root_path=self.recipe_root,
            step_name="split",
            relative_path="train.parquet",
        )
        train_df = pd.read_parquet(train_data_path)
        validate_classification_config(self.task, self.positive_class, train_df, self.target_col)

        validation_data_path = get_step_output_path(
            recipe_root_path=self.recipe_root,
            step_name="split",
            relative_path="validation.parquet",
        )
        validation_df = pd.read_parquet(validation_data_path)

        sys.path.append(self.recipe_root)

        def get_identity_transformer():
            from sklearn.pipeline import Pipeline
            from sklearn.preprocessing import FunctionTransformer

            return Pipeline(steps=[("identity", FunctionTransformer())])

        if "transformer_method" not in self.step_config and self.step_config["using"] == "custom":
            raise MlflowException(
                "Missing 'transformer_method' configuration in the transform step, "
                "which is using 'custom'.",
                error_code=INVALID_PARAMETER_VALUE,
            )
        method_config = self.step_config.get("transformer_method")
        transformer = None
        if method_config and self.step_config["using"] == "custom":
            transformer_fn = getattr(
                importlib.import_module(_USER_DEFINED_TRANSFORM_STEP_MODULE), method_config
            )
            transformer = _validate_user_code_output(transformer_fn)
        transformer = transformer if transformer else get_identity_transformer()
        transformer.fit(train_df.drop(columns=[self.target_col]), train_df[self.target_col])

        def transform_dataset(dataset):
            features = dataset.drop(columns=[self.target_col])
            transformed_features = transformer.transform(features)
            if not isinstance(transformed_features, pd.DataFrame):
                num_features = transformed_features.shape[1]
                columns = _get_output_feature_names(transformer, num_features, features.columns)
                transformed_features = pd.DataFrame(transformed_features, columns=columns)
            transformed_features[self.target_col] = dataset[self.target_col].values
            return transformed_features

        train_transformed = transform_dataset(train_df)
        validation_transformed = transform_dataset(validation_df)

        with open(os.path.join(output_directory, "transformer.pkl"), "wb") as f:
            cloudpickle.dump(transformer, f)

        train_transformed.to_parquet(
            os.path.join(output_directory, "transformed_training_data.parquet")
        )
        validation_transformed.to_parquet(
            os.path.join(output_directory, "transformed_validation_data.parquet")
        )

        self.run_end_time = time.time()
        self.execution_duration = self.run_end_time - run_start_time

        return self._build_profiles_and_card(train_df, train_transformed, transformer)

    def _build_profiles_and_card(self, train_df, train_transformed, transformer) -> BaseCard:
        # Build card
        card = BaseCard(self.recipe_name, self.name)

        if not self.skip_data_profiling:
            # Tab 1: build profiles for train_transformed
            train_transformed_profile = get_pandas_data_profiles(
                [["Profile of Train Transformed Dataset", train_transformed]]
            )
            card.add_tab("Data Profile (Train Transformed)", "{{PROFILE}}").add_pandas_profile(
                "PROFILE", train_transformed_profile
            )

        # Tab 3: transformer diagram
        from sklearn import set_config
        from sklearn.utils import estimator_html_repr

        set_config(display="diagram")
        transformer_repr = estimator_html_repr(transformer)
        card.add_tab("Transformer", "{{TRANSFORMER}}").add_html("TRANSFORMER", transformer_repr)

        # Tab 4: transformer input schema
        card.add_tab("Input Schema", "{{INPUT_SCHEMA}}").add_html(
            "INPUT_SCHEMA",
            BaseCard.render_table({"Name": n, "Type": t} for n, t in train_df.dtypes.items()),
        )

        # Tab 5: transformer output schema
        try:
            card.add_tab("Output Schema", "{{OUTPUT_SCHEMA}}").add_html(
                "OUTPUT_SCHEMA",
                BaseCard.render_table(
                    {"Name": n, "Type": t} for n, t in train_transformed.dtypes.items()
                ),
            )
        except Exception as e:
            card.add_tab("Output Schema", "{{OUTPUT_SCHEMA}}").add_html(
                "OUTPUT_SCHEMA", f"Failed to extract transformer schema. Error: {e}"
            )

        # Tab 6: transformer output data preview
        card.add_tab("Data Preview", "{{DATA_PREVIEW}}").add_html(
            "DATA_PREVIEW", BaseCard.render_table(train_transformed.head())
        )

        # Tab 7: run summary
        (
            card.add_tab(
                "Run Summary",
                """
                {{ EXE_DURATION }}
                {{ LAST_UPDATE_TIME }}
                """,
            )
        )

        return card

    @classmethod
    def from_recipe_config(cls, recipe_config, recipe_root):
        step_config = {}
        if recipe_config.get("steps", {}).get("transform", {}) is not None:
            step_config.update(recipe_config.get("steps", {}).get("transform", {}))
        step_config["target_col"] = recipe_config.get("target_col")
        step_config["recipe"] = recipe_config.get("recipe", "regression/v1")
        if "positive_class" in recipe_config:
            step_config["positive_class"] = recipe_config.get("positive_class")
        step_config.update(
            get_recipe_tracking_config(
                recipe_root_path=recipe_root,
                recipe_config=recipe_config,
            ).to_dict()
        )
        return cls(step_config, recipe_root)

    @property
    def name(self):
        return "transform"

    def get_artifacts(self):
        return [
            DataframeArtifact(
                "transformed_training_data",
                self.recipe_root,
                self.name,
                "transformed_training_data.parquet",
            ),
            DataframeArtifact(
                "transformed_validation_data",
                self.recipe_root,
                self.name,
                "transformed_validation_data.parquet",
            ),
            TransformerArtifact(
                "transformer", self.recipe_root, self.name, self.tracking_config.tracking_uri
            ),
        ]

    def step_class(self):
        return StepClass.TRAINING



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: databricks_filesystem_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import databricks_pb2 as databricks__pb2
from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#databricks_filesystem_service.proto\x12\x11mlflow.filesystem\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\")\n\nHttpHeader\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"`\n\x18\x43reateDownloadUrlRequest\x12\x0c\n\x04path\x18\x01 \x01(\t:6\xe2?3\n1com.databricks.rpc.RPC[CreateDownloadUrlResponse]\"\x82\x01\n\x19\x43reateDownloadUrlResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12.\n\x07headers\x18\x02 \x03(\x0b\x32\x1d.mlflow.filesystem.HttpHeader:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\\\n\x16\x43reateUploadUrlRequest\x12\x0c\n\x04path\x18\x01 \x01(\t:4\xe2?1\n/com.databricks.rpc.RPC[CreateUploadUrlResponse]\"\x80\x01\n\x17\x43reateUploadUrlResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12.\n\x07headers\x18\x02 \x03(\x0b\x32\x1d.mlflow.filesystem.HttpHeader:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\x96\x01\n\x0e\x44irectoryEntry\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x14\n\x0cis_directory\x18\x02 \x01(\x08\x12\x11\n\tfile_size\x18\x03 \x01(\x03\x12\x15\n\rlast_modified\x18\x04 \x01(\x03\x12\x0c\n\x04name\x18\x05 \x01(\t:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\x8f\x01\n\x15ListDirectoryResponse\x12\x33\n\x08\x63ontents\x18\x01 \x03(\x0b\x32!.mlflow.filesystem.DirectoryEntry\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:(\xe2?%\n#com.databricks.rpc.DoNotLogContents2\xcb\x02\n\x11\x46ilesystemService\x12\x9d\x01\n\x11\x43reateDownloadUrl\x12+.mlflow.filesystem.CreateDownloadUrlRequest\x1a,.mlflow.filesystem.CreateDownloadUrlResponse\"-\xf2\x86\x19)\n%\n\x04POST\x12\x17/fs/create-download-url\x1a\x04\x08\x02\x10\x00\x10\x03\x12\x95\x01\n\x0f\x43reateUploadUrl\x12).mlflow.filesystem.CreateUploadUrlRequest\x1a*.mlflow.filesystem.CreateUploadUrlResponse\"+\xf2\x86\x19\'\n#\n\x04POST\x12\x15/fs/create-upload-url\x1a\x04\x08\x02\x10\x00\x10\x03\x42\x30\n#com.databricks.api.proto.filesystem\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')



_HTTPHEADER = DESCRIPTOR.message_types_by_name['HttpHeader']
_CREATEDOWNLOADURLREQUEST = DESCRIPTOR.message_types_by_name['CreateDownloadUrlRequest']
_CREATEDOWNLOADURLRESPONSE = DESCRIPTOR.message_types_by_name['CreateDownloadUrlResponse']
_CREATEUPLOADURLREQUEST = DESCRIPTOR.message_types_by_name['CreateUploadUrlRequest']
_CREATEUPLOADURLRESPONSE = DESCRIPTOR.message_types_by_name['CreateUploadUrlResponse']
_DIRECTORYENTRY = DESCRIPTOR.message_types_by_name['DirectoryEntry']
_LISTDIRECTORYRESPONSE = DESCRIPTOR.message_types_by_name['ListDirectoryResponse']
HttpHeader = _reflection.GeneratedProtocolMessageType('HttpHeader', (_message.Message,), {
  'DESCRIPTOR' : _HTTPHEADER,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.HttpHeader)
  })
_sym_db.RegisterMessage(HttpHeader)

CreateDownloadUrlRequest = _reflection.GeneratedProtocolMessageType('CreateDownloadUrlRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDOWNLOADURLREQUEST,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateDownloadUrlRequest)
  })
_sym_db.RegisterMessage(CreateDownloadUrlRequest)

CreateDownloadUrlResponse = _reflection.GeneratedProtocolMessageType('CreateDownloadUrlResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDOWNLOADURLRESPONSE,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateDownloadUrlResponse)
  })
_sym_db.RegisterMessage(CreateDownloadUrlResponse)

CreateUploadUrlRequest = _reflection.GeneratedProtocolMessageType('CreateUploadUrlRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEUPLOADURLREQUEST,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateUploadUrlRequest)
  })
_sym_db.RegisterMessage(CreateUploadUrlRequest)

CreateUploadUrlResponse = _reflection.GeneratedProtocolMessageType('CreateUploadUrlResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEUPLOADURLRESPONSE,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateUploadUrlResponse)
  })
_sym_db.RegisterMessage(CreateUploadUrlResponse)

DirectoryEntry = _reflection.GeneratedProtocolMessageType('DirectoryEntry', (_message.Message,), {
  'DESCRIPTOR' : _DIRECTORYENTRY,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.DirectoryEntry)
  })
_sym_db.RegisterMessage(DirectoryEntry)

ListDirectoryResponse = _reflection.GeneratedProtocolMessageType('ListDirectoryResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTDIRECTORYRESPONSE,
  '__module__' : 'databricks_filesystem_service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.filesystem.ListDirectoryResponse)
  })
_sym_db.RegisterMessage(ListDirectoryResponse)

_FILESYSTEMSERVICE = DESCRIPTOR.services_by_name['FilesystemService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#com.databricks.api.proto.filesystem\220\001\001\240\001\001\342?\002\020\001'
  _CREATEDOWNLOADURLREQUEST._options = None
  _CREATEDOWNLOADURLREQUEST._serialized_options = b'\342?3\n1com.databricks.rpc.RPC[CreateDownloadUrlResponse]'
  _CREATEDOWNLOADURLRESPONSE._options = None
  _CREATEDOWNLOADURLRESPONSE._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
  _CREATEUPLOADURLREQUEST._options = None
  _CREATEUPLOADURLREQUEST._serialized_options = b'\342?1\n/com.databricks.rpc.RPC[CreateUploadUrlResponse]'
  _CREATEUPLOADURLRESPONSE._options = None
  _CREATEUPLOADURLRESPONSE._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
  _DIRECTORYENTRY._options = None
  _DIRECTORYENTRY._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
  _LISTDIRECTORYRESPONSE._options = None
  _LISTDIRECTORYRESPONSE._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
  _FILESYSTEMSERVICE.methods_by_name['CreateDownloadUrl']._options = None
  _FILESYSTEMSERVICE.methods_by_name['CreateDownloadUrl']._serialized_options = b'\362\206\031)\n%\n\004POST\022\027/fs/create-download-url\032\004\010\002\020\000\020\003'
  _FILESYSTEMSERVICE.methods_by_name['CreateUploadUrl']._options = None
  _FILESYSTEMSERVICE.methods_by_name['CreateUploadUrl']._serialized_options = b'\362\206\031\'\n#\n\004POST\022\025/fs/create-upload-url\032\004\010\002\020\000\020\003'
  _HTTPHEADER._serialized_start=99
  _HTTPHEADER._serialized_end=140
  _CREATEDOWNLOADURLREQUEST._serialized_start=142
  _CREATEDOWNLOADURLREQUEST._serialized_end=238
  _CREATEDOWNLOADURLRESPONSE._serialized_start=241
  _CREATEDOWNLOADURLRESPONSE._serialized_end=371
  _CREATEUPLOADURLREQUEST._serialized_start=373
  _CREATEUPLOADURLREQUEST._serialized_end=465
  _CREATEUPLOADURLRESPONSE._serialized_start=468
  _CREATEUPLOADURLRESPONSE._serialized_end=596
  _DIRECTORYENTRY._serialized_start=599
  _DIRECTORYENTRY._serialized_end=749
  _LISTDIRECTORYRESPONSE._serialized_start=752
  _LISTDIRECTORYRESPONSE._serialized_end=895
  _FILESYSTEMSERVICE._serialized_start=898
  _FILESYSTEMSERVICE._serialized_end=1229
FilesystemService = service_reflection.GeneratedServiceType('FilesystemService', (_service.Service,), dict(
  DESCRIPTOR = _FILESYSTEMSERVICE,
  __module__ = 'databricks_filesystem_service_pb2'
  ))

FilesystemService_Stub = service_reflection.GeneratedServiceStubType('FilesystemService_Stub', (FilesystemService,), dict(
  DESCRIPTOR = _FILESYSTEMSERVICE,
  __module__ = 'databricks_filesystem_service_pb2'
  ))


# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import base64
import html
import logging
import os
import pathlib
import pickle
import random
import re
import string
from io import StringIO
from typing import Optional, Union

from packaging.version import Version

from mlflow.exceptions import INVALID_PARAMETER_VALUE, MlflowException

CARD_PICKLE_NAME = "card.pkl"
CARD_HTML_NAME = "card.html"

_PP_VARIABLE_LINK_REGEX = re.compile(r'<a\s+href="?(?P<href>#pp_var_[-0-9]+)"?\s*>')


_logger = logging.getLogger(__name__)


class CardTab:
    def __init__(self, name: str, template: str) -> None:
        """
        Construct a step card tab with supported HTML template.

        Args:
            name: A string representing the name of the tab.
            template: A string representing the HTML template for the card content.
        """
        import jinja2
        from jinja2 import meta as jinja2_meta

        self.name = name
        self.template = template

        j2_env = jinja2.Environment()
        self._variables = jinja2_meta.find_undeclared_variables(j2_env.parse(template))
        self._context = {}

    def add_html(self, name: str, html_content: str) -> CardTab:
        """
        Adds html to the CardTab.

        Args:
            name: Name of the variable in the Jinja2 template.
            html_content: The html to replace the named template variable.

        Returns:
            The updated card instance.
        """
        if name not in self._variables:
            raise MlflowException(
                f"{name} is not a valid template variable defined in template: '{self.template}'",
                error_code=INVALID_PARAMETER_VALUE,
            )
        self._context[name] = html_content
        return self

    def add_markdown(self, name: str, markdown: str) -> CardTab:
        """
        Adds markdown to the card replacing the variable name in the CardTab template.

        Args:
            name: Name of the variable in the CardTab Jinja2 template.
            markdown: The markdown content.

        Returns:
            The updated card tab instance.
        """
        from markdown import markdown as md_to_html

        self.add_html(name, md_to_html(markdown))
        return self

    def add_image(
        self,
        name: str,
        image_file_path: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> None:
        if not os.path.exists(image_file_path):
            self.add_html(name, "Image Unavailable")
            _logger.warning(f"Unable to locate image file {image_file_path} to render {name}.")
            return

        with open(image_file_path, "rb") as f:
            base64_str = base64.b64encode(f.read()).decode("utf-8")

        image_type = pathlib.Path(image_file_path).suffix[1:]

        width_style = f'width="{width}"' if width else ""
        height_style = f'height="{width}"' if height else ""
        img_html = (
            f'<img src="data:image/{image_type};base64, {base64_str}" '
            f"{width_style} {height_style} />"
        )
        self.add_html(name, img_html)

    def add_pandas_profile(self, name: str, profile: str) -> CardTab:
        """
        Add a new tab representing the provided pandas profile to the card.

        Args:
            name: Name of the variable in the Jinja2 template.
            profile: HTML string to render profile in the step card.

        Returns:
            The updated card instance.
        """
        try:
            profile_iframe = (
                "<iframe srcdoc='{src}' width='100%' height='500' frameborder='0'></iframe>"
            ).format(src=html.escape(profile))
        except Exception as e:
            profile_iframe = f"Unable to create data profile. Error found:\n{e}"
        self.add_html(name, profile_iframe)
        return self

    def to_html(self) -> str:
        """
        Returns a rendered HTML representing the content of the tab.

        Returns:
            a HTML string
        """
        from jinja2 import BaseLoader
        from jinja2.sandbox import SandboxedEnvironment

        j2_env = SandboxedEnvironment(loader=BaseLoader()).from_string(self.template)
        return j2_env.render({**self._context})


class BaseCard:
    def __init__(self, recipe_name: str, step_name: str) -> None:
        """
        BaseCard Constructor

        Args:
            recipe_name: A string representing name of the recipe.
            step_name: A string representing the name of the step.
        """
        self._recipe_name = recipe_name
        self._step_name = step_name
        self._template_name = "base.html"

        self._string_builder = StringIO()
        self._tabs = []

    def add_tab(self, name, html_template) -> CardTab:
        """
        Add a new tab with arbitrary content.

        Args:
            name: A string representing the name of the tab.
            html_template: A string representing the HTML template for the card content.
        """
        tab = CardTab(name, html_template)
        self._tabs.append((name, tab))
        return tab

    def get_tab(self, name) -> Union[CardTab, None]:
        """
        Returns an existing tab with the specified name. Returns None if not found.

        Args:
            name: A string representing the name of the tab.

        Returns:
            An existing tab with the specified name. If not found, returns None.
        """
        for key, tab in self._tabs:
            if key == name:
                return tab
        return None

    def add_text(self, text: str) -> BaseCard:
        """
        Add text to the textual representation of this card.

        Args:
            text: A string text.

        Returns:
            The updated card instance.
        """
        self._string_builder.write(text)
        return self

    def to_html(self) -> str:
        """
        This function renders the Jinja2 template based on the provided context so far.

        Returns:
            A HTML string.
        """
        import jinja2

        def get_random_id(length=6):
            return "".join(
                random.choice(string.ascii_lowercase + string.digits) for _ in range(length)
            )

        base_template_path = os.path.join(os.path.dirname(__file__), "templates")
        j2_env = jinja2.Environment(loader=jinja2.FileSystemLoader(base_template_path))
        tab_list = [(name, tab.to_html()) for name, tab in self._tabs]
        page_id = get_random_id()
        return j2_env.get_template(self._template_name).render(
            {
                "HEADER_TITLE": f"{self._step_name.capitalize()}@{self._recipe_name}",
                "TABLINK": f"tablink-{page_id}",
                "CONTENT": f"content-{page_id}",
                "BUTTON_CONTAINER": f"button-container-{page_id}",
                "tab_list": tab_list,
            }
        )

    def to_text(self) -> str:
        """
        Returns:
            The textual representation of the card.
        """
        return self._string_builder.getvalue()

    def save_as_html(self, path) -> None:
        if os.path.isdir(path):
            path = os.path.join(path, CARD_HTML_NAME)
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.to_html())

    def save(self, path: str) -> None:
        if os.path.isdir(path):
            path = os.path.join(path, CARD_PICKLE_NAME)
        with open(path, "wb") as out:
            pickle.dump(self, out)

    @staticmethod
    def load(path):
        if os.path.isdir(path):
            path = os.path.join(path, CARD_PICKLE_NAME)
        with open(path, "rb") as f:
            return pickle.load(f)

    @staticmethod
    def render_table(table, columns=None, hide_index=True):
        """
        Renders a table-like object as an HTML table.

        Args:
            table: Table-like object (e.g. pandas DataFrame, 2D numpy array, list of tuples).
            columns: Column names to use. If `table` doesn't have column names, this argument
                provides names for the columns. Otherwise, only the specified columns will be
                included in the output HTML table.
            hide_index: Hide index column when rendering.
        """
        import pandas as pd
        from pandas.io.formats.style import Styler

        pandas_version = Version(pd.__version__)

        if not isinstance(table, Styler):
            table = pd.DataFrame(table, columns=columns)
            # Escape specific characters in HTML to prevent
            # javascript code injection
            # Note that `pandas_df.style.to_html(escape=True) does not work
            # So that we have to manually escape values in dataframe cells.

            def escape_value(x):
                return html.escape(str(x))

            if hasattr(table, "map"):
                table = table.map(escape_value)
            else:
                if pandas_version >= Version("2.1.0"):
                    table = table.map(escape_value)
                else:
                    table = table.applymap(escape_value)
            table = table.style

        styler = table.set_table_attributes('style="border-collapse:collapse"').set_table_styles(
            [
                {
                    "selector": "table, th, td",
                    "props": [
                        ("border", "1px solid grey"),
                        ("text-align", "left"),
                        ("padding", "5px"),
                    ],
                },
            ]
        )
        if hide_index:
            rendered_table = (
                styler.hide(axis="index").to_html()
                if pandas_version >= Version("1.4.0")
                else styler.hide_index().render()
            )
        else:
            rendered_table = (
                styler.to_html() if pandas_version >= Version("1.4.0") else styler.render()
            )
        return f'<div style="max-height: 500px; overflow: scroll;">{rendered_table}</div>'


class FailureCard(BaseCard):
    """
    Step card providing information about a failed step execution, including a stacktrace.

    TODO: Migrate the failure card to a tab-based card, removing this class and its associated
          HTML template in the process.
    """

    def __init__(
        self, recipe_name: str, step_name: str, failure_traceback: str, output_directory: str
    ):
        super().__init__(
            recipe_name=recipe_name,
            step_name=step_name,
        )
        self.add_tab("Step Status", "{{ STEP_STATUS }}").add_html(
            "STEP_STATUS",
            '<p><strong>Step status: <span style="color:red">Failed</span></strong></p>',
        )
        self.add_tab(
            "Stacktrace",
            (
                "<div class='stacktrace-container'><p style='margin-top:0px'><code>"
                "{{ STACKTRACE|e }}</code></p></div>"
            ),
        ).add_html("STACKTRACE", str(failure_traceback))
        warning_output_path = os.path.join(output_directory, "warning_logs.txt")
        if os.path.exists(warning_output_path):
            with open(warning_output_path) as f:
                self.add_tab("Warning Logs", "{{ STEP_WARNINGS }}").add_html(
                    "STEP_WARNINGS", f"<pre>{f.read()}</pre>"
                )



import time
from dataclasses import asdict, dataclass, field
from typing import List, Literal, Optional

from mlflow.types.schema import Array, ColSpec, DataType, Object, Property, Schema

# TODO: Switch to pydantic in a future version of MLflow.
#       For now, to prevent adding pydantic as a core dependency,
#       we use dataclasses instead.
#
#       Unfortunately, validation for generic types is not that
#       straightforward. For example, `isinstance(thing, List[T])``
#       is not supported, so the code here is a little ugly.


class _BaseDataclass:
    def _validate_field(self, key, val_type, required):
        value = getattr(self, key, None)
        if required and value is None:
            raise ValueError(f"`{key}` is required")
        if value is not None and not isinstance(value, val_type):
            raise ValueError(
                f"`{key}` must be of type {val_type.__name__}, got {type(value).__name__}"
            )

    def _validate_list(self, key, val_type, required):
        values = getattr(self, key, None)
        if required and values is None:
            raise ValueError(f"`{key}` is required")

        if values is not None:
            if isinstance(values, list) and not all(isinstance(v, val_type) for v in values):
                raise ValueError(f"All items in `{key}` must be of type {val_type.__name__}")
            elif not isinstance(values, list):
                raise ValueError(f"`{key}` must be a list, got {type(values).__name__}")

    def _convert_dataclass_list(self, key, cls):
        values = getattr(self, key)
        if not isinstance(values, list):
            raise ValueError(f"`{key}` must be a list")

        if len(values) > 0:
            # if the items are all dicts, try to convert them to the desired class
            if all(isinstance(v, dict) for v in values):
                try:
                    setattr(self, key, [cls(**v) for v in values])
                except TypeError as e:
                    raise ValueError(f"Error when coercing {values} to {cls.__name__}: {e}")
            elif any(not isinstance(v, cls) for v in values):
                raise ValueError(
                    f"Items in `{key}` must all have the same type: {cls.__name__} or dict"
                )

    def to_dict(self):
        return asdict(self, dict_factory=lambda obj: {k: v for (k, v) in obj if v is not None})


@dataclass
class ChatMessage(_BaseDataclass):
    """
    A message in a chat request or response.

    Args:
        role (str): The role of the entity that sent the message (e.g. ``"user"``, ``"system"``).
        content (str): The content of the message.
        name (str): The name of the entity that sent the message. **Optional**.
    """

    role: str
    content: str
    name: Optional[str] = None

    def __post_init__(self):
        self._validate_field("role", str, True)
        self._validate_field("content", str, True)
        self._validate_field("name", str, False)


@dataclass
class ChatParams(_BaseDataclass):
    """
    Common parameters used for chat inference

    Args:
        temperature (float): A param used to control randomness and creativity during inference.
            **Optional**, defaults to ``1.0``
        max_tokens (int): The maximum number of new tokens to generate.
            **Optional**, defaults to ``None`` (unlimited)
        stop (List[str]): A list of tokens at which to stop generation.
            **Optional**, defaults to ``None``
        n (int): The number of responses to generate.
            **Optional**, defaults to ``1``
        stream (bool): Whether to stream back responses as they are generated.
            **Optional**, defaults to ``False``
    """

    temperature: float = 1.0
    max_tokens: Optional[int] = None
    stop: Optional[List[str]] = None
    n: int = 1
    stream: bool = False

    def __post_init__(self):
        self._validate_field("temperature", float, True)
        self._validate_field("max_tokens", int, False)
        self._validate_list("stop", str, False)
        self._validate_field("n", int, True)
        self._validate_field("stream", bool, True)


@dataclass()
class ChatRequest(ChatParams):
    """
    Format of the request object expected by the chat endpoint.

    Args:
        messages (List[:py:class:`ChatMessage`]): A list of :py:class:`ChatMessage`
            that will be passed to the model. **Optional**, defaults to empty list (``[]``)
        temperature (float): A param used to control randomness and creativity during inference.
            **Optional**, defaults to ``1.0``
        max_tokens (int): The maximum number of new tokens to generate.
            **Optional**, defaults to ``None`` (unlimited)
        stop (List[str]): A list of tokens at which to stop generation. **Optional**,
            defaults to ``None``
        n (int): The number of responses to generate. **Optional**,
            defaults to ``1``
        stream (bool): Whether to stream back responses as they are generated. **Optional**,
            defaults to ``False``
    """

    messages: List[ChatMessage] = field(default_factory=list)

    def __post_init__(self):
        self._convert_dataclass_list("messages", ChatMessage)
        super().__post_init__()


@dataclass
class ChatChoice(_BaseDataclass):
    """
    A single chat response generated by the model.

    Args:
        index (int): The index of the response in the list of responses.
        message (:py:class:`ChatMessage`): The message that was generated.
        finish_reason (str): The reason why generation stopped.
    """

    index: int
    message: ChatMessage
    finish_reason: str

    def __post_init__(self):
        self._validate_field("index", int, True)
        self._validate_field("finish_reason", str, True)
        if isinstance(self.message, dict):
            self.message = ChatMessage(**self.message)
        if not isinstance(self.message, ChatMessage):
            raise ValueError(
                f"Expected `message` to be of type ChatMessage or dict, got {type(self.message)}"
            )


@dataclass
class TokenUsageStats(_BaseDataclass):
    """
    Stats about the number of tokens used during inference.

    Args:
        prompt_tokens (int): The number of tokens in the prompt.
        completion_tokens (int): The number of tokens in the generated completion.
        total_tokens (int): The total number of tokens used.
    """

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    def __post_init__(self):
        self._validate_field("prompt_tokens", int, True)
        self._validate_field("completion_tokens", int, True)
        self._validate_field("total_tokens", int, True)


@dataclass
class ChatResponse(_BaseDataclass):
    """
    The full response object returned by the chat endpoint.

    Args:
        id (str): The ID of the response.
        object (str): The object type.
        created (int): The time the response was created.
            **Optional**, defaults to the current time.
        model (str): The name of the model used.
        choices (List[:py:class:`ChatChoice`]): A list of :py:class:`ChatChoice` objects
            containing the generated responses
        usage (:py:class:`TokenUsageStats`): An object describing the tokens used by the request.
    """

    id: str
    model: str
    choices: List[ChatChoice]
    usage: TokenUsageStats
    object: Literal["chat.completion"] = "chat.completion"
    created: int = field(default_factory=lambda: int(time.time()))

    def __post_init__(self):
        self._validate_field("id", str, True)
        self._validate_field("object", str, True)
        self._validate_field("created", int, True)
        self._validate_field("model", str, True)
        self._convert_dataclass_list("choices", ChatChoice)
        if isinstance(self.usage, dict):
            self.usage = TokenUsageStats(**self.usage)
        if not isinstance(self.usage, TokenUsageStats):
            raise ValueError(
                f"Expected `usage` to be of type TokenUsageStats or dict, got {type(self.usage)}"
            )


CHAT_MODEL_INPUT_SCHEMA = Schema(
    [
        ColSpec(
            name="messages",
            type=Array(
                Object(
                    [
                        Property("role", DataType.string),
                        Property("content", DataType.string),
                        Property("name", DataType.string, False),
                    ]
                )
            ),
        ),
        ColSpec(name="temperature", type=DataType.double, required=False),
        ColSpec(name="max_tokens", type=DataType.long, required=False),
        ColSpec(name="stop", type=Array(DataType.string), required=False),
        ColSpec(name="n", type=DataType.long, required=False),
        ColSpec(name="stream", type=DataType.boolean, required=False),
    ]
)

CHAT_MODEL_OUTPUT_SCHEMA = Schema(
    [
        ColSpec(name="id", type=DataType.string),
        ColSpec(name="object", type=DataType.string),
        ColSpec(name="created", type=DataType.long),
        ColSpec(name="model", type=DataType.string),
        ColSpec(
            name="choices",
            type=Array(
                Object(
                    [
                        Property("index", DataType.long),
                        Property(
                            "message",
                            Object(
                                [
                                    Property("role", DataType.string),
                                    Property("content", DataType.string),
                                    Property("name", DataType.string, False),
                                ]
                            ),
                        ),
                        Property("finish_reason", DataType.string),
                    ]
                )
            ),
        ),
        ColSpec(
            name="usage",
            type=Object(
                [
                    Property("prompt_tokens", DataType.long),
                    Property("completion_tokens", DataType.long),
                    Property("total_tokens", DataType.long),
                ]
            ),
        ),
    ]
)

CHAT_MODEL_INPUT_EXAMPLE = {
    "messages": [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Hello!"},
    ],
    "temperature": 1.0,
    "max_tokens": 10,
    "stop": ["\n"],
    "n": 1,
    "stream": False,
}

COMPLETIONS_MODEL_INPUT_SCHEMA = Schema(
    [
        ColSpec(name="prompt", type=DataType.string),
        ColSpec(name="temperature", type=DataType.double, required=False),
        ColSpec(name="max_tokens", type=DataType.long, required=False),
        ColSpec(name="stop", type=Array(DataType.string), required=False),
        ColSpec(name="n", type=DataType.long, required=False),
        ColSpec(name="stream", type=DataType.boolean, required=False),
    ]
)

COMPLETIONS_MODEL_OUTPUT_SCHEMA = Schema(
    [
        ColSpec(name="id", type=DataType.string),
        ColSpec(name="object", type=DataType.string),
        ColSpec(name="created", type=DataType.long),
        ColSpec(name="model", type=DataType.string),
        ColSpec(
            name="choices",
            type=Array(
                Object(
                    [
                        Property("index", DataType.long),
                        Property(
                            "text",
                            DataType.string,
                        ),
                        Property("finish_reason", DataType.string),
                    ]
                )
            ),
        ),
        ColSpec(
            name="usage",
            type=Object(
                [
                    Property("prompt_tokens", DataType.long),
                    Property("completion_tokens", DataType.long),
                    Property("total_tokens", DataType.long),
                ]
            ),
        ),
    ]
)

EMBEDDING_MODEL_INPUT_SCHEMA = Schema(
    [
        ColSpec(name="input", type=DataType.string),
    ]
)

EMBEDDING_MODEL_OUTPUT_SCHEMA = Schema(
    [
        ColSpec(name="object", type=DataType.string),
        ColSpec(
            name="data",
            type=Array(
                Object(
                    [
                        Property("index", DataType.long),
                        Property("object", DataType.string),
                        Property("embedding", Array(DataType.double)),
                    ]
                )
            ),
        ),
        ColSpec(
            name="usage",
            type=Object(
                [
                    Property("prompt_tokens", DataType.long),
                    Property("total_tokens", DataType.long),
                ]
            ),
        ),
    ]
)



from abc import abstractmethod
from typing import Any, Dict

from mlflow.data.dataset_source import DatasetSource


class FileSystemDatasetSource(DatasetSource):
    """
    Represents the source of a dataset stored on a filesystem, e.g. a local UNIX filesystem,
    blob storage services like S3, etc.
    """

    @property
    @abstractmethod
    def uri(self):
        """The URI referring to the dataset source filesystem location.

        Returns:
            The URI referring to the dataset source filesystem location,
            e.g "s3://mybucket/path/to/mydataset", "/tmp/path/to/my/dataset" etc.

        """

    @staticmethod
    @abstractmethod
    def _get_source_type() -> str:
        """
        Returns:
            A string describing the filesystem containing the dataset, e.g. "local", "s3", ...
        """

    @abstractmethod
    def load(self, dst_path=None) -> str:
        """Downloads the dataset source to the local filesystem.

        Args:
            dst_path: Path of the local filesystem destination directory to which to download the
                dataset source. If the directory does not exist, it is created. If
                unspecified, the dataset source is downloaded to a new uniquely-named
                directory on the local filesystem, unless the dataset source already
                exists on the local filesystem, in which case its local path is returned
                directly.

        Returns:
            The path to the downloaded dataset source on the local filesystem.

        """

    @staticmethod
    @abstractmethod
    def _can_resolve(raw_source: Any) -> bool:
        """
        Args:
            raw_source: The raw source, e.g. a string like "s3://mybucket/path/to/iris/data".

        Returns:
            True if this DatsetSource can resolve the raw source, False otherwise.
        """

    @classmethod
    @abstractmethod
    def _resolve(cls, raw_source: Any) -> "FileSystemDatasetSource":
        """
        Args:
            raw_source: The raw source, e.g. a string like "s3://mybucket/path/to/iris/data".
        """

    @abstractmethod
    def to_dict(self) -> Dict[Any, Any]:
        """
        Returns:
            A JSON-compatible dictionary representation of the FileSystemDatasetSource.
        """

    @classmethod
    @abstractmethod
    def from_dict(cls, source_dict: Dict[Any, Any]) -> "FileSystemDatasetSource":
        """
        Args:
            source_dict: A dictionary representation of the FileSystemDatasetSource.
        """



import logging

from mlflow.environment_variables import (
    MLFLOW_HUGGINGFACE_DISABLE_ACCELERATE_FEATURES,
    MLFLOW_HUGGINGFACE_MODEL_MAX_SHARD_SIZE,
)
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_STATE
from mlflow.transformers.flavor_config import FlavorKey, get_peft_base_model, is_peft_model

_logger = logging.getLogger(__name__)

# File/directory names for saved artifacts
_MODEL_BINARY_FILE_NAME = "model"
_COMPONENTS_BINARY_DIR_NAME = "components"
_PROCESSOR_BINARY_DIR_NAME = "processor"


def save_pipeline_pretrained_weights(path, pipeline, flavor_conf, processor=None):
    """
    Save the binary artifacts of the pipeline to the specified local path.

    Args:
        path: The local path to save the pipeline
        pipeline: Transformers pipeline instance
        flavor_config: The flavor configuration constructed for the pipeline
        processor: Optional processor instance to save alongside the pipeline
    """
    model = get_peft_base_model(pipeline.model) if is_peft_model(pipeline.model) else pipeline.model

    model.save_pretrained(
        save_directory=path.joinpath(_MODEL_BINARY_FILE_NAME),
        max_shard_size=MLFLOW_HUGGINGFACE_MODEL_MAX_SHARD_SIZE.get(),
    )

    component_dir = path.joinpath(_COMPONENTS_BINARY_DIR_NAME)
    for name in flavor_conf.get(FlavorKey.COMPONENTS, []):
        getattr(pipeline, name).save_pretrained(component_dir.joinpath(name))

    if processor:
        processor.save_pretrained(component_dir.joinpath(_PROCESSOR_BINARY_DIR_NAME))


def load_model_and_components_from_local(path, flavor_conf, accelerate_conf, device=None):
    """
    Load the model and components of a Transformer pipeline from the specified local path.

    Args:
        path: The local path contains MLflow model artifacts
        flavor_conf: The flavor configuration
        accelerate_conf: The configuration for the accelerate library
        device: The device to load the model onto
    """
    loaded = {}

    # NB: Path resolution for models that were saved prior to 2.4.1 release when the patching for
    #     the saved pipeline or component artifacts was handled by duplicate entries for components
    #     (artifacts/pipeline/* and artifacts/components/*) and pipelines were saved via the
    #     "artifacts/pipeline/*" path. In order to load the older formats after the change, the
    #     presence of the new path key is checked.
    model_path = path.joinpath(flavor_conf.get(FlavorKey.MODEL_BINARY, "pipeline"))
    loaded[FlavorKey.MODEL] = _load_model(model_path, flavor_conf, accelerate_conf, device)

    components = flavor_conf.get(FlavorKey.COMPONENTS, [])
    if FlavorKey.PROCESSOR_TYPE in flavor_conf:
        components.append("processor")

    for component_key in components:
        loaded[component_key] = _load_component(flavor_conf, component_key, local_path=path)

    return loaded


def load_model_and_components_from_huggingface_hub(flavor_conf, accelerate_conf, device=None):
    """
    Load the model and components of a Transformer pipeline from HuggingFace Hub.

    Args:
        flavor_conf: The flavor configuration
        accelerate_conf: The configuration for the accelerate library
        device: The device to load the model onto
    """
    loaded = {}

    model_repo = flavor_conf[FlavorKey.MODEL_NAME]
    model_revision = flavor_conf.get(FlavorKey.MODEL_REVISION)

    if not model_revision:
        raise MlflowException(
            "The model was saved with 'save_pretrained' set to False, but the commit hash is not "
            "found in the saved metadata. Loading the model with the different version may cause "
            "inconsistency issue and security risk.",
            error_code=INVALID_STATE,
        )

    loaded[FlavorKey.MODEL] = _load_model(
        model_repo, flavor_conf, accelerate_conf, device, revision=model_revision
    )

    components = flavor_conf.get(FlavorKey.COMPONENTS, [])
    if FlavorKey.PROCESSOR_TYPE in flavor_conf:
        components.append("processor")

    for name in components:
        loaded[name] = _load_component(flavor_conf, name)

    return loaded


def _load_component(flavor_conf, name, local_path=None):
    import transformers

    _COMPONENT_TO_AUTOCLASS_MAP = {
        FlavorKey.TOKENIZER: transformers.AutoTokenizer,
        FlavorKey.FEATURE_EXTRACTOR: transformers.AutoFeatureExtractor,
        FlavorKey.PROCESSOR: transformers.AutoProcessor,
        FlavorKey.IMAGE_PROCESSOR: transformers.AutoImageProcessor,
    }

    component_name = flavor_conf[FlavorKey.COMPONENT_TYPE.format(name)]
    if hasattr(transformers, component_name):
        cls = getattr(transformers, component_name)
        trust_remote = False
    else:
        if local_path is None:
            raise MlflowException(
                f"A custom component `{component_name}` was specified, "
                "but no local config file was found to retrieve the "
                "definition. Make sure your model was saved with "
                "save_pretrained=True."
            )
        cls = _COMPONENT_TO_AUTOCLASS_MAP[name]
        trust_remote = True

    if local_path is not None:
        # Load component from local file
        path = local_path.joinpath(_COMPONENTS_BINARY_DIR_NAME, name)
        return cls.from_pretrained(str(path), trust_remote_code=trust_remote)
    else:
        # Load component from HuggingFace Hub
        repo = flavor_conf[FlavorKey.COMPONENT_NAME.format(name)]
        revision = flavor_conf.get(FlavorKey.COMPONENT_REVISION.format(name))
        return cls.from_pretrained(repo, revision=revision, trust_remote_code=trust_remote)


def _load_class_from_transformers_config(model_name_or_path, revision=None):
    """
    This method retrieves the Transformers AutoClass from the transformers config.
    Using the correct AutoClass allows us to leverage Transformers' model loading
    machinery, which is necessary for supporting models using custom code.
    """
    import transformers
    from transformers import AutoConfig

    config = AutoConfig.from_pretrained(
        model_name_or_path,
        revision=revision,
        # trust_remote_code is set to True in order to
        # make sure the config gets loaded as the correct
        # class. if this is not set for custom models, the
        # base class will be loaded instead of the custom one.
        trust_remote_code=True,
    )

    # the model's class name (e.g. "MPTForCausalLM")
    # is stored in the `architectures` field. it
    # seems to usually just have one element.
    class_name = config.architectures[0]

    # if the class is available in transformers natively,
    # then we don't need to execute any custom code.
    if hasattr(transformers, class_name):
        cls = getattr(transformers, class_name)
        return cls, False
    else:
        # else, we need to fetch the correct AutoClass.
        # this is defined in the `auto_map` field. there
        # should only be one AutoClass that maps to the
        # model's class name.
        auto_classes = [
            auto_class
            for auto_class, module in config.auto_map.items()
            if module.split(".")[-1] == class_name
        ]

        if len(auto_classes) == 0:
            raise MlflowException(f"Couldn't find a loader class for {class_name}")

        auto_class = auto_classes[0]
        cls = getattr(transformers, auto_class)

        # we will need to trust remote code when loading the model
        return cls, True


def _load_model(model_name_or_path, flavor_conf, accelerate_conf, device, revision=None):
    """
    Try to load a model with various loading strategies.
      1. Try to load the model with accelerate
      2. Try to load the model with the specified device
      3. Load the model without the device
    """
    import transformers

    if hasattr(transformers, flavor_conf[FlavorKey.MODEL_TYPE]):
        cls = getattr(transformers, flavor_conf[FlavorKey.MODEL_TYPE])
        trust_remote = False
    else:
        cls, trust_remote = _load_class_from_transformers_config(
            model_name_or_path, revision=revision
        )

    load_kwargs = {"revision": revision} if revision else {}
    if trust_remote:
        load_kwargs.update({"trust_remote_code": True})

    if model := _try_load_model_with_accelerate(
        cls, model_name_or_path, {**accelerate_conf, **load_kwargs}
    ):
        return model

    load_kwargs["device"] = device
    if torch_dtype := flavor_conf.get(FlavorKey.TORCH_DTYPE):
        load_kwargs[FlavorKey.TORCH_DTYPE] = torch_dtype

    if model := _try_load_model_with_device(cls, model_name_or_path, load_kwargs):
        return model
    _logger.warning(
        "Could not specify device parameter for this pipeline type."
        "Falling back to loading the model with the default device."
    )

    load_kwargs.pop("device", None)
    return cls.from_pretrained(model_name_or_path, **load_kwargs)


def _try_load_model_with_accelerate(model_class, model_name_or_path, load_kwargs):
    if MLFLOW_HUGGINGFACE_DISABLE_ACCELERATE_FEATURES.get():
        return None

    try:
        return model_class.from_pretrained(model_name_or_path, **load_kwargs)
    except (ValueError, TypeError, NotImplementedError, ImportError):
        # NB: ImportError is caught here in the event that `accelerate` is not installed
        # on the system, which will raise if `low_cpu_mem_usage` is set or the argument
        # `device_map` is set and accelerate is not installed.
        pass


def _try_load_model_with_device(model_class, model_name_or_path, load_kwargs):
    try:
        return model_class.from_pretrained(model_name_or_path, **load_kwargs)
    except OSError as e:
        revision = load_kwargs.get("revision")
        if f"{revision} is not a valid git identifier" in str(e):
            raise MlflowException(
                f"The model was saved with a HuggingFace Hub repository name '{model_name_or_path}'"
                f"and a commit hash '{revision}', but the commit is not found in the repository. "
            )
        else:
            raise e
    except (ValueError, TypeError, NotImplementedError):
        pass



import json
import os
import posixpath

from mlflow.entities import FileInfo
from mlflow.protos.databricks_artifacts_pb2 import ArtifactCredentialInfo
from mlflow.protos.databricks_filesystem_service_pb2 import (
    CreateDownloadUrlRequest,
    CreateDownloadUrlResponse,
    CreateUploadUrlRequest,
    CreateUploadUrlResponse,
    FilesystemService,
    ListDirectoryResponse,
)
from mlflow.store.artifact.cloud_artifact_repo import CloudArtifactRepository
from mlflow.utils.file_utils import download_file_using_http_uri
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.request_utils import augmented_raise_for_status, cloud_storage_http_request
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    call_endpoint,
    extract_api_info_for_service,
)

FILESYSTEM_METHOD_TO_INFO = extract_api_info_for_service(FilesystemService, _REST_API_PATH_PREFIX)
DIRECTORIES_ENDPOINT = "/api/2.0/fs/directories"


class PresignedUrlArtifactRepository(CloudArtifactRepository):
    """
    Stores and retrieves model artifacts using presigned URLs.
    """

    def __init__(self, db_creds, model_full_name, model_version):
        artifact_uri = posixpath.join(
            "/Models", model_full_name.replace(".", "/"), str(model_version)
        )
        super().__init__(artifact_uri)
        self.db_creds = db_creds

    def log_artifact(self, local_file, artifact_path=None):
        artifact_file_path = os.path.basename(local_file)
        if artifact_path:
            artifact_file_path = posixpath.join(artifact_path, artifact_file_path)
        cloud_credentials = self._get_write_credential_infos(
            remote_file_paths=[artifact_file_path]
        )[0]
        self._upload_to_cloud(
            cloud_credential_info=cloud_credentials,
            src_file_path=local_file,
            artifact_file_path=None,
        )

    def _get_write_credential_infos(self, remote_file_paths):
        endpoint, method = FILESYSTEM_METHOD_TO_INFO[CreateUploadUrlRequest]
        credential_infos = []
        for relative_path in remote_file_paths:
            fs_full_path = posixpath.join(self.artifact_uri, relative_path)
            req_body = message_to_json(CreateUploadUrlRequest(path=fs_full_path))
            response_proto = CreateUploadUrlResponse()
            resp = call_endpoint(
                host_creds=self.db_creds,
                endpoint=endpoint,
                method=method,
                json_body=req_body,
                response_proto=response_proto,
            )
            headers = [
                ArtifactCredentialInfo.HttpHeader(name=header.name, value=header.value)
                for header in resp.headers
            ]
            credential_infos.append(ArtifactCredentialInfo(signed_uri=resp.url, headers=headers))
        return credential_infos

    def _upload_to_cloud(self, cloud_credential_info, src_file_path, artifact_file_path=None):
        # artifact_file_path is unused in this implementation because the presigned URL
        # and local file path are sufficient for upload to cloud storage
        presigned_url = cloud_credential_info.signed_uri
        headers = {header.name: header.value for header in cloud_credential_info.headers}
        with open(src_file_path, "rb") as source_file:
            data = source_file.read()
            with cloud_storage_http_request(
                "put", presigned_url, data=data, headers=headers
            ) as response:
                augmented_raise_for_status(response)

    def list_artifacts(self, path=""):
        infos = []
        page_token = ""
        while True:
            endpoint = posixpath.join(DIRECTORIES_ENDPOINT, self.artifact_uri.lstrip("/"), path)
            req_body = json.dumps({"page_token": page_token}) if page_token else ""

            response_proto = ListDirectoryResponse()
            resp = call_endpoint(
                host_creds=self.db_creds,
                endpoint=endpoint,
                method="GET",
                json_body=req_body,
                response_proto=response_proto,
            )
            for dir_entry in resp.contents:
                rel_path = posixpath.relpath(dir_entry.path, self.artifact_uri)
                if dir_entry.is_directory:
                    infos.append(FileInfo(rel_path, True, None))
                else:
                    infos.append(FileInfo(rel_path, False, dir_entry.file_size))
            page_token = resp.next_page_token
            if not page_token:
                break
        return sorted(infos, key=lambda f: f.path)

    def _get_read_credential_infos(self, remote_file_paths):
        credential_infos = []
        for remote_file_path in remote_file_paths:
            resp = self._get_download_presigned_url_and_headers(remote_file_path)
            headers = [
                ArtifactCredentialInfo.HttpHeader(name=header.name, value=header.value)
                for header in resp.headers
            ]
            credential_infos.append(ArtifactCredentialInfo(signed_uri=resp.url, headers=headers))
        return credential_infos

    def _download_from_cloud(self, remote_file_path, local_path):
        resp = self._get_download_presigned_url_and_headers(remote_file_path)
        presigned_url = resp.url
        headers = {header.name: header.value for header in resp.headers}
        download_file_using_http_uri(
            http_uri=presigned_url, download_path=local_path, headers=headers
        )

    def _get_download_presigned_url_and_headers(self, remote_file_path):
        remote_file_full_path = posixpath.join(self.artifact_uri, remote_file_path)
        endpoint, method = FILESYSTEM_METHOD_TO_INFO[CreateDownloadUrlRequest]
        req_body = message_to_json(CreateDownloadUrlRequest(path=remote_file_full_path))
        response_proto = CreateDownloadUrlResponse()
        return call_endpoint(
            host_creds=self.db_creds,
            endpoint=endpoint,
            method=method,
            json_body=req_body,
            response_proto=response_proto,
        )



import os
import platform
import shutil
import subprocess
import sys

import yaml

import mlflow
from mlflow import MlflowClient
from mlflow.environment_variables import MLFLOW_WHEELED_MODEL_PIP_DOWNLOAD_OPTIONS
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import BAD_REQUEST
from mlflow.pyfunc.model import MLMODEL_FILE_NAME, Model
from mlflow.store.artifact.utils.models import _parse_model_uri, get_model_name_and_version
from mlflow.tracking.artifact_utils import _download_artifact_from_uri
from mlflow.utils.annotations import experimental
from mlflow.utils.environment import (
    _REQUIREMENTS_FILE_NAME,
    _get_pip_deps,
    _mlflow_additional_pip_env,
    _overwrite_pip_deps,
)
from mlflow.utils.model_utils import _validate_and_prepare_target_save_path
from mlflow.utils.uri import get_databricks_profile_uri_from_artifact_uri

_WHEELS_FOLDER_NAME = "wheels"
_ORIGINAL_REQ_FILE_NAME = "original_requirements.txt"
_PLATFORM = "platform"


@experimental
class WheeledModel:
    """
    Helper class to create a model with added dependency wheels from an existing registered model.
    The `wheeled` model contains all the model dependencies as wheels stored as model artifacts.
    .. note::
        This utility only operates on a model that has been registered to the Model Registry.
    """

    def __init__(self, model_uri):
        self._model_uri = model_uri
        databricks_profile_uri = (
            get_databricks_profile_uri_from_artifact_uri(model_uri) or mlflow.get_registry_uri()
        )
        client = MlflowClient(registry_uri=databricks_profile_uri)
        self._model_name, _ = get_model_name_and_version(client, model_uri)

    @classmethod
    def log_model(cls, model_uri, registered_model_name=None):
        """
        Logs a registered model as an MLflow artifact for the current run. This only operates on
        a model which has been registered to the Model Registry. Given a registered model_uri (
        e.g. models:/<model_name>/<model_version>), this utility re-logs the model along with all
        the required model libraries back to the Model Registry. The required model libraries are
        stored along with the model as model artifacts. In addition, supporting files to the
        model (e.g. conda.yaml, requirements.txt) are modified to use the added libraries.

        By default, this utility creates a new model version under the same registered model
        specified by ``model_uri``. This behavior can be overridden by specifying the
        ``registered_model_name`` argument.

        Args:
            model_uri: A registered model uri in the Model Registry of the form
                       models:/<model_name>/<model_version/stage/latest>
            registered_model_name: The new model version (model with its libraries) is
                                   registered under the inputted registered_model_name. If None,
                                   a new version is logged to the existing model in the Model
                                   Registry.

        .. code-block:: python
            :caption: Example

            # Given a model uri, log the wheeled model
            with mlflow.start_run():
                WheeledModel.log_model(model_uri)
        """
        parsed_uri = _parse_model_uri(model_uri)
        return Model.log(
            artifact_path=None,
            flavor=WheeledModel(model_uri),
            registered_model_name=registered_model_name or parsed_uri.name,
        )

    def save_model(self, path, mlflow_model=None):
        """
        Given an existing registered model, saves the model along with it's dependencies stored as
        wheels to a path on the local file system.

        This does not modify existing model behavior or existing model flavors. It simply downloads
        the model dependencies as wheels and modifies the requirements.txt and conda.yaml file to
        point to the downloaded wheels.

        The download_command defaults to downloading only binary packages using the
        `--only-binary=:all:` option. This behavior can be overridden using an environment
        variable `MLFLOW_WHEELED_MODEL_PIP_DOWNLOAD_OPTIONS`, which will allows setting
        different options such as `--prefer-binary`, `--no-binary`, etc.

        Args:
            path: Local path where the model is to be saved.
            mlflow_model: The new :py:mod:`mlflow.models.Model` metadata file to store the
                updated model metadata.
        """
        from mlflow.pyfunc import ENV, FLAVOR_NAME, _extract_conda_env

        path = os.path.abspath(path)
        _validate_and_prepare_target_save_path(path)

        local_model_path = _download_artifact_from_uri(self._model_uri, output_path=path)

        wheels_dir = os.path.join(local_model_path, _WHEELS_FOLDER_NAME)
        pip_requirements_path = os.path.join(local_model_path, _REQUIREMENTS_FILE_NAME)
        model_metadata_path = os.path.join(local_model_path, MLMODEL_FILE_NAME)

        model_metadata = Model.load(model_metadata_path)

        # Check if the model file has `wheels` set to True
        if model_metadata.__dict__.get(_WHEELS_FOLDER_NAME, None) is not None:
            raise MlflowException("Model libraries are already added", BAD_REQUEST)

        conda_env = _extract_conda_env(model_metadata.flavors.get(FLAVOR_NAME, {}).get(ENV, None))
        conda_env_path = os.path.join(local_model_path, conda_env)
        if conda_env is None and not os.path.isfile(pip_requirements_path):
            raise MlflowException(
                "Cannot add libraries for model with no logged dependencies.", BAD_REQUEST
            )

        if not os.path.isfile(pip_requirements_path):
            self._create_pip_requirement(conda_env_path, pip_requirements_path)

        WheeledModel._download_wheels(
            pip_requirements_path=pip_requirements_path, dst_path=wheels_dir
        )

        # Keep a copy of the original requirement.txt
        shutil.copy2(pip_requirements_path, os.path.join(local_model_path, _ORIGINAL_REQ_FILE_NAME))

        # Update requirements.txt with wheels
        pip_deps = self._overwrite_pip_requirements_with_wheels(
            pip_requirements_path=pip_requirements_path, wheels_dir=wheels_dir
        )

        # Update conda.yaml with wheels
        self._update_conda_env(pip_deps, conda_env_path)

        # Update MLModel File
        mlflow_model = self._update_mlflow_model(
            original_model_metadata=model_metadata, mlflow_model=mlflow_model
        )
        mlflow_model.save(model_metadata_path)
        return mlflow_model

    def _update_conda_env(self, new_pip_deps, conda_env_path):
        """
        Updates the list pip packages in the conda.yaml file to the list of wheels in the wheels
        directory.
        {
            "name": "env",
            "channels": [...],
            "dependencies": [
                ...,
                "pip",
                {"pip": [...]},  <- Overwrite this with list of wheels
            ],
        }

        Args:
            new_pip_deps: List of pip dependencies as wheels
            conda_env_path: Path to conda.yaml file in the model directory
        """
        with open(conda_env_path) as f:
            conda_env = yaml.safe_load(f)

        new_conda_env = _overwrite_pip_deps(conda_env, new_pip_deps)

        with open(conda_env_path, "w") as out:
            yaml.safe_dump(new_conda_env, stream=out, default_flow_style=False)

    def _update_mlflow_model(self, original_model_metadata, mlflow_model):
        """
        Modifies the MLModel file to reflect updated information such as the run_id,
        utc_time_created. Additionally, this also adds `wheels` to the MLModel file to indicate that
        this is a `wheeled` model.

        Args:
            original_model_file_path: The model metadata stored in the original MLmodel file.
            mlflow_model: :py:mod:`mlflow.models.Model` configuration of the newly created
                          wheeled model
        """

        run_id = mlflow.tracking.fluent._get_or_start_run().info.run_id
        if mlflow_model is None:
            mlflow_model = Model(run_id=run_id)

        original_model_metadata.__dict__.update(
            {k: v for k, v in mlflow_model.__dict__.items() if v}
        )
        mlflow_model.__dict__.update(original_model_metadata.__dict__)
        mlflow_model.artifact_path = WheeledModel.get_wheel_artifact_path(
            mlflow_model.artifact_path
        )

        mlflow_model.wheels = {_PLATFORM: platform.platform()}
        return mlflow_model

    @classmethod
    def _download_wheels(cls, pip_requirements_path, dst_path):
        """
        Downloads all the wheels of the dependencies specified in the requirements.txt file.
        The pip wheel download_command defaults to downloading only binary packages using
        the `--only-binary=:all:` option. This behavior can be overridden using an
        environment variable `MLFLOW_WHEELED_MODEL_PIP_DOWNLOAD_OPTIONS`, which will allows
        setting different options such as `--prefer-binary`, `--no-binary`, etc.

        Args:
            pip_requirements_path: Path to requirements.txt in the model directory
            dst_path: Path to the directory where the wheels are to be downloaded
        """
        if not os.path.exists(dst_path):
            os.makedirs(dst_path)

        pip_wheel_options = MLFLOW_WHEELED_MODEL_PIP_DOWNLOAD_OPTIONS.get()

        try:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "wheel",
                    pip_wheel_options,
                    "--wheel-dir",
                    dst_path,
                    "-r",
                    pip_requirements_path,
                    "--no-cache-dir",
                ],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            raise MlflowException(
                f"An error occurred while downloading the dependency wheels: {e.stderr}"
            )

    def _overwrite_pip_requirements_with_wheels(self, pip_requirements_path, wheels_dir):
        """
        Overwrites the requirements.txt with the wheels of the required dependencies.

        Args:
            pip_requirements_path: Path to requirements.txt in the model directory.
            wheels_dir: Path to directory where wheels are stored.
        """
        wheels = []
        with open(pip_requirements_path, "w") as wheels_requirements:
            for wheel_file in os.listdir(wheels_dir):
                if wheel_file.endswith(".whl"):
                    complete_wheel_file = os.path.join(_WHEELS_FOLDER_NAME, wheel_file)
                    wheels.append(complete_wheel_file)
                    wheels_requirements.write(complete_wheel_file + "\n")
        return wheels

    def _create_pip_requirement(self, conda_env_path, pip_requirements_path):
        """
        This method creates a requirements.txt file for the model dependencies if the file does not
        already exist. It uses the pip dependencies found in the conda.yaml env file.

        Args:
            conda_env_path: Path to conda.yaml env file which contains the required pip
                dependencies
            pip_requirements_path: Path where the new requirements.txt will be created.
        """
        with open(conda_env_path) as f:
            conda_env = yaml.safe_load(f)
        pip_deps = _get_pip_deps(conda_env)
        _mlflow_additional_pip_env(pip_deps, pip_requirements_path)

    @classmethod
    def get_wheel_artifact_path(cls, original_artifact_path):
        return original_artifact_path + "_" + _WHEELS_FOLDER_NAME
